from __future__ import annotations
from typing import Any, TYPE_CHECKING, overload
from collections.abc import Iterable, Iterator
import logging
from typing_extensions import Self

import archinfo
import claripy

from angr.knowledge_plugins.key_definitions.environment import Environment
from angr.knowledge_plugins.key_definitions.tag import Tag
from angr.knowledge_plugins.key_definitions.heap_address import HeapAddress
from angr.knowledge_plugins.key_definitions.definition import A
from angr.engines.light import SpOffset
from angr.code_location import CodeLocation
from angr.storage.memory_mixins.paged_memory.pages.multi_values import MVType, MultiValues
from angr.storage.memory_mixins import MultiValuedMemory
from angr.knowledge_plugins.key_definitions import LiveDefinitions, DerefSize, Definition
from angr.knowledge_plugins.key_definitions.atoms import Atom, GuardUse, Register, MemoryLocation, ConstantSrc
from .heap_allocator import HeapAllocator
from .subject import Subject, SubjectType
from .rd_initializer import RDAStateInitializer

if TYPE_CHECKING:
    from .reaching_definitions import ReachingDefinitionsAnalysis


l = logging.getLogger(name=__name__)

#
# Reaching definitions state
#


class ReachingDefinitionsState:
    """
    Represents the internal state of the ReachingDefinitionsAnalysis.

    It contains a data class LiveDefinitions, which stores both definitions and uses for register, stack, memory, and
    temporary variables, uncovered during the analysis.

    :param subject: The subject being analyzed.
    :ivar arch: The architecture targeted by the program.
    :param track_tmps: Only tells whether or not temporary variables should be taken into consideration when
                              representing the state of the analysis.
                              Should be set to true when the analysis has counted uses and definitions for temporary
                              variables, false otherwise.
    :param analysis: The analysis that generated the state represented by this object.
    :param rtoc_value: When the targeted architecture is ppc64, the initial function needs to know the `rtoc_value`.
    :param live_definitions:
    :param canonical_size:
        The sizes (in bytes) that objects with an UNKNOWN_SIZE are treated as for operations where sizes are necessary.
    :param heap_allocator: Mechanism to model the management of heap memory.
    :param environment: Representation of the environment of the analyzed program.
    """

    __slots__ = (
        "_canonical_size",
        "_element_limit",
        "_environment",
        "_sp_adjusted",
        "_subject",
        "_track_consts",
        "_track_tmps",
        "all_definitions",
        "analysis",
        "arch",
        "codeloc",
        "codeloc_uses",
        "exit_observed",
        "heap_allocator",
        "live_definitions",
    )

    def __init__(
        self,
        codeloc: CodeLocation,
        arch: archinfo.Arch,
        subject: Subject,
        analysis: ReachingDefinitionsAnalysis,
        track_tmps: bool = False,
        track_consts: bool = False,
        rtoc_value=None,
        live_definitions: LiveDefinitions | None = None,
        canonical_size: int = 8,
        heap_allocator: HeapAllocator | None = None,
        environment: Environment | None = None,
        sp_adjusted: bool = False,
        all_definitions: set[Definition[A]] | None = None,
        initializer: RDAStateInitializer | None = None,
        element_limit: int = 5,
        merge_into_tops: bool = True,
    ):
        # handy short-hands
        self.codeloc = codeloc
        self.arch: archinfo.Arch = arch
        self._subject = subject
        self._track_tmps = track_tmps
        self._track_consts = track_consts
        self.analysis = analysis
        self._canonical_size: int = canonical_size
        self._sp_adjusted: bool = sp_adjusted
        self._element_limit: int = element_limit

        self.all_definitions: set[Definition[Any]] = set() if all_definitions is None else all_definitions

        self.heap_allocator = heap_allocator or HeapAllocator(canonical_size)
        self._environment: Environment = environment or Environment()

        self.codeloc_uses: set[Definition[Any]] = set()

        # have we observed an exit statement or not during the analysis of the *last instruction* of a block? we should
        # not perform any sp updates if it is the case. this is for handling conditional returns in ARM binaries.
        # this variable is not copied to new states because it only tracks if an exit statement is observed in a single
        # block and is always set to False at the beginning of the analysis of each block.
        self.exit_observed: bool = False

        # initialize the live definitions
        # This must stay at the end of the __init__ method, because the _set_initialization_values method will call
        # the state initializer which might need to access some of the above attributes, e.g. the heap_allocator
        # to do its job

        if live_definitions is None:
            # the first time this state is created. initialize it
            self.live_definitions = LiveDefinitions(
                self.arch,
                track_tmps=self._track_tmps,
                canonical_size=canonical_size,
                element_limit=element_limit,
                merge_into_tops=merge_into_tops,
            )
            if self.analysis is not None:
                self.live_definitions.project = self.analysis.project
            self._set_initialization_values(
                subject, rtoc_value, initializer=initializer, project=self.live_definitions.project
            )
        else:
            # this state is a copy from a previous state. skip the initialization
            self.live_definitions = live_definitions

    #
    # Util methods for working with the memory model
    #

    def top(self, bits: int):
        return self.live_definitions.top(bits)

    def is_top(self, *args):
        return self.live_definitions.is_top(*args)

    def heap_address(self, offset: int | HeapAddress) -> claripy.ast.BV:
        return self.live_definitions.heap_address(offset)

    @staticmethod
    def is_heap_address(addr: claripy.ast.Base) -> bool:
        return LiveDefinitions.is_heap_address(addr)

    @staticmethod
    def get_heap_offset(addr: claripy.ast.Base) -> int | None:
        return LiveDefinitions.get_heap_offset(addr)

    def stack_address(self, offset: int) -> claripy.ast.BV:
        return self.live_definitions.stack_address(offset)

    def is_stack_address(self, addr: claripy.ast.Base) -> bool:
        return self.live_definitions.is_stack_address(addr)

    def get_stack_offset(self, addr: claripy.ast.Base) -> int | None:
        offset = self.live_definitions.get_stack_offset(addr)
        if offset is not None:
            return self._to_signed(offset)
        return None

    def _initial_stack_pointer(self):
        if self.arch.bits == 32:
            return claripy.BVS("stack_base", 32, explicit_name=True)
        if self.arch.bits == 64:
            return claripy.BVS("stack_base", 64, explicit_name=True)
        raise ValueError(f"Unsupported architecture word size {self.arch.bits}")

    def _to_signed(self, n):
        if n >= 2 ** (self.arch.bits - 1):
            # convert it to a negative number
            return n - 2**self.arch.bits
        return n

    def annotate_with_def(self, symvar: MVType, definition: Definition[Any]) -> MVType:
        """

        :param symvar:
        :param definition:
        :return:
        """
        return self.live_definitions.annotate_with_def(symvar, definition)

    def annotate_mv_with_def(self, mv: MultiValues[MVType], definition: Definition[A]) -> MultiValues[MVType]:
        return MultiValues(
            offset_to_values={
                offset: {self.annotate_with_def(value, definition) for value in values} for offset, values in mv.items()
            }
        )

    def extract_defs(self, symvar: claripy.ast.Base) -> Iterator[Definition[Any]]:
        yield from self.live_definitions.extract_defs(symvar)

    #
    # Other methods
    #

    @property
    def tmps(self):
        return self.live_definitions.tmps

    @property
    def tmp_uses(self):
        return self.live_definitions.tmp_uses

    @property
    def registers(self) -> MultiValuedMemory:
        return self.live_definitions.registers

    @property
    def register_uses(self):
        return self.live_definitions.register_uses

    @property
    def stack(self) -> MultiValuedMemory:
        return self.live_definitions.stack

    @property
    def stack_uses(self):
        return self.live_definitions.stack_uses

    @property
    def heap(self) -> MultiValuedMemory:
        return self.live_definitions.heap

    @property
    def heap_uses(self):
        return self.live_definitions.heap_uses

    @property
    def memory(self) -> MultiValuedMemory:
        return self.live_definitions.memory

    @property
    def memory_uses(self):
        return self.live_definitions.memory_uses

    @property
    def others(self) -> dict[Atom, MultiValues]:
        return self.live_definitions.others

    @property
    def uses_by_codeloc(self):
        return self.live_definitions.uses_by_codeloc

    def get_sp(self) -> int:
        return self.live_definitions.get_sp()

    def get_stack_address(self, offset: claripy.ast.Base) -> int | None:
        return self.live_definitions.get_stack_address(offset)

    @property
    def environment(self):
        return self._environment

    @property
    def _dep_graph(self):
        return self.analysis._dep_graph

    @property
    def dep_graph(self):
        return self.analysis.dep_graph

    def __repr__(self):
        ctnt = f"RDState-{self.live_definitions!r}"
        return f"{{{ctnt}}}"

    def _set_initialization_values(
        self,
        subject: Subject,
        rtoc_value: int | None = None,
        initializer: RDAStateInitializer | None = None,
        project=None,
    ):
        if initializer is None:
            initializer = RDAStateInitializer(self.arch, project=project)

        if subject.type == SubjectType.Function:
            if isinstance(self.arch, archinfo.arch_ppc64.ArchPPC64) and not rtoc_value:
                raise ValueError("The architecture being ppc64, the parameter `rtoc_value` should be provided.")

            initializer.initialize_function_state(self, subject.cc, subject.content.addr, rtoc_value)
        elif subject.type == SubjectType.CallTrace:
            if isinstance(self.arch, archinfo.arch_ppc64.ArchPPC64) and not rtoc_value:
                raise ValueError("The architecture being ppc64, the parameter `rtoc_value` should be provided.")

            initializer.initialize_function_state(
                self, subject.cc, subject.content.current_function_address(), rtoc_value
            )
        elif subject.type == SubjectType.Block:
            pass

        return self

    def copy(self, discard_tmpdefs=False) -> Self:
        return type(self)(
            self.codeloc,
            self.arch,
            self._subject,
            track_tmps=self._track_tmps,
            track_consts=self._track_consts,
            analysis=self.analysis,
            live_definitions=self.live_definitions.copy(discard_tmpdefs=discard_tmpdefs),
            canonical_size=self._canonical_size,
            heap_allocator=self.heap_allocator,
            environment=self._environment,
            sp_adjusted=self._sp_adjusted,
            all_definitions=self.all_definitions.copy(),
            element_limit=self._element_limit,
        )

    def merge(self, *others: Self) -> tuple[Self, bool]:
        state = self.copy()

        state.live_definitions, merged_0 = state.live_definitions.merge(*[other.live_definitions for other in others])
        state._environment, merged_1 = state.environment.merge(*[other.environment for other in others])

        return state, merged_0 or merged_1

    def compare(self, other: ReachingDefinitionsState) -> bool:
        r0 = self.live_definitions.compare(other.live_definitions)
        r1 = self.environment.compare(other.environment)

        return r0 and r1

    def move_codelocs(self, new_codeloc: CodeLocation) -> None:
        if self.codeloc != new_codeloc:
            self.codeloc = new_codeloc
            self.codeloc_uses = set()

    def kill_definitions(self, atom: Atom) -> None:
        """
        Overwrite existing definitions w.r.t 'atom' with a dummy definition instance. A dummy definition will not be
        removed during simplification.
        """
        existing_defs = self.live_definitions.get_definitions(atom)

        self.live_definitions.kill_definitions(atom)

        for def_ in existing_defs:
            self.analysis.model.kill_def(def_)

    def kill_and_add_definition(
        self,
        atom: Atom,
        data: MultiValues,
        dummy=False,
        tags: set[Tag] | None = None,
        endness=None,  # XXX destroy
        annotated: bool = False,
        uses: set[Definition[A]] | None = None,
        override_codeloc: CodeLocation | None = None,
    ) -> tuple[MultiValues | None, set[Definition[A]]]:
        codeloc = override_codeloc or self.codeloc
        existing_defs = self.live_definitions.get_definitions(atom)
        mv = self.live_definitions.kill_and_add_definition(
            atom, codeloc, data, dummy=dummy, tags=tags, endness=endness, annotated=annotated
        )

        if mv is not None:
            defs = set(LiveDefinitions.extract_defs_from_mv(mv))
            self.all_definitions |= defs

            if self._dep_graph is not None:
                stack_use = {u for u in self.codeloc_uses if isinstance(u.atom, MemoryLocation) and u.atom.is_on_stack}

                sp_offset = self.arch.sp_offset
                bp_offset = self.arch.bp_offset

                values = set()
                for vs in mv.values():
                    for v in vs:
                        values.add(v)

                if uses is None:
                    uses = self.codeloc_uses
                for used in uses:
                    # sp is always used as a stack pointer, and we do not track dependencies against stack pointers.
                    # bp is sometimes used as a base pointer. we recognize such cases by checking if there is a use to
                    # the stack variable.
                    #
                    # There are two cases for which it is superfluous to report a dependency on (a use of) stack/base
                    # pointers:
                    # - The `Definition` *uses* a `MemoryLocation` pointing to the stack;
                    # - The `Definition` *is* a `MemoryLocation` pointing to the stack.
                    is_using_spbp_while_memory_address_on_stack_is_used = (
                        isinstance(used.atom, Register)
                        and used.atom.reg_offset in (sp_offset, bp_offset)
                        and len(stack_use) > 0
                    )
                    is_using_spbp_to_define_memory_location_on_stack = (
                        isinstance(atom, MemoryLocation)
                        and (
                            atom.is_on_stack
                            or (isinstance(atom.addr, claripy.ast.Base) and self.is_stack_address(atom.addr))
                        )
                        and isinstance(used.atom, Register)
                        and used.atom.reg_offset in (sp_offset, bp_offset)
                    )

                    if not (
                        is_using_spbp_while_memory_address_on_stack_is_used
                        or is_using_spbp_to_define_memory_location_on_stack
                    ):
                        # Moderately confusing misnomers. This is an edge from a def to a use, since the
                        # "uses" are actually the definitions that we're using and the "definition" is the
                        # new definition; i.e. The def that the old def is used to construct so this is
                        # really a graph where nodes are defs and edges are uses.
                        self._dep_graph.add_node(used)
                        for def_ in defs:
                            if not def_.dummy:
                                self._dep_graph.add_edge(used, def_)

                        cfg = self.analysis.project.kb.cfgs.get_most_accurate()
                        self._dep_graph.add_dependencies_for_concrete_pointers_of(
                            values,
                            used,
                            cfg,
                            self.analysis.project.loader,
                        )
        else:
            defs = set()

        for def_ in existing_defs:
            self.analysis.model.kill_def(def_)
        for def_ in defs:
            self.analysis.model.add_def(def_)

        return mv, defs

    def add_use(self, atom: Atom, expr: Any | None = None) -> None:
        self.codeloc_uses.update(self.get_definitions(atom))
        self.live_definitions.add_use(atom, self.codeloc, expr=expr)

    def add_use_by_def(self, definition: Definition[A], expr: Any | None = None) -> None:
        self.codeloc_uses.add(definition)
        self.live_definitions.add_use_by_def(definition, self.codeloc, expr=expr)

    def add_tmp_use(self, tmp: int, expr: Any | None = None) -> None:
        defs = self.live_definitions.get_tmp_definitions(tmp)
        self.add_tmp_use_by_defs(defs, expr=expr)

    def add_tmp_use_by_defs(self, defs: Iterable[Definition[A]], expr: Any | None = None) -> None:  # pylint:disable=unused-argument
        for definition in defs:
            self.codeloc_uses.add(definition)
            # if track_tmps is False, definitions may not be Tmp definitions
            self.live_definitions.add_use_by_def(definition, self.codeloc, expr=expr)

    def add_register_use(self, reg_offset: int, size: int, expr: Any | None = None) -> None:
        defs = self.live_definitions.get_register_definitions(reg_offset, size)
        self.add_register_use_by_defs(defs, expr=expr)

    def add_register_use_by_defs(self, defs: Iterable[Definition[A]], expr: Any | None = None) -> None:
        for definition in defs:
            self.codeloc_uses.add(definition)
            self.live_definitions.add_register_use_by_def(definition, self.codeloc, expr=expr)

    def add_stack_use(self, stack_offset: int, size: int, expr: Any | None = None) -> None:
        defs = self.live_definitions.get_stack_definitions(stack_offset, size)
        self.add_stack_use_by_defs(defs, expr=expr)

    def add_stack_use_by_defs(self, defs: Iterable[Definition[A]], expr: Any | None = None):
        for definition in defs:
            self.codeloc_uses.add(definition)
            self.live_definitions.add_stack_use_by_def(definition, self.codeloc, expr=expr)

    def add_heap_use(self, heap_offset: int, size: int, expr: Any | None = None) -> None:
        defs = self.live_definitions.get_heap_definitions(heap_offset, size)
        self.add_heap_use_by_defs(defs, expr=expr)

    def add_heap_use_by_defs(self, defs: Iterable[Definition[A]], expr: Any | None = None):
        for definition in defs:
            self.codeloc_uses.add(definition)
            self.live_definitions.add_heap_use_by_def(definition, self.codeloc, expr=expr)

    def add_memory_use_by_def(self, definition: Definition[A], expr: Any | None = None):
        self.codeloc_uses.add(definition)
        self.live_definitions.add_memory_use_by_def(definition, self.codeloc, expr=expr)

    def add_memory_use_by_defs(self, defs: Iterable[Definition[A]], expr: Any | None = None):
        for definition in defs:
            self.codeloc_uses.add(definition)
            self.live_definitions.add_memory_use_by_def(definition, self.codeloc, expr=expr)

    def get_definitions(
        self, atom: Atom | Definition[Atom] | Iterable[Atom] | Iterable[Definition[Atom]] | MultiValues
    ) -> set[Definition[Atom]]:
        return self.live_definitions.get_definitions(atom)

    def get_values(self, spec: A | Definition[A] | Iterable[A]) -> MultiValues | None:
        return self.live_definitions.get_values(spec)

    def get_one_value(
        self, spec: A | Definition[A] | Iterable[A] | Iterable[Definition[A]], strip_annotations: bool = False
    ) -> claripy.ast.bv.BV | None:
        return self.live_definitions.get_one_value(spec, strip_annotations=strip_annotations)

    @overload
    def get_concrete_value(self, spec: Atom | Definition[Atom] | Iterable[Atom], cast_to: type[int]) -> int | None: ...

    @overload
    def get_concrete_value(
        self, spec: Atom | Definition[Atom] | Iterable[Atom], cast_to: type[bytes]
    ) -> bytes | None: ...

    def get_concrete_value(
        self, spec: Atom | Definition[Atom] | Iterable[Atom], cast_to: type[int] | type[bytes] = int
    ) -> int | bytes | None:
        return self.live_definitions.get_concrete_value(spec, cast_to)

    def mark_guard(self, target):
        atom = GuardUse(target)
        kinda_definition = Definition(atom, self.codeloc)

        if self._dep_graph is not None:
            self._dep_graph.add_node(kinda_definition)
            for used in self.codeloc_uses:
                self._dep_graph.add_edge(used, kinda_definition)

    def mark_const(self, value: int, size: int):
        atom = ConstantSrc(value, size)
        kinda_definition = Definition(atom, self.codeloc)

        if self._dep_graph is not None and self._track_consts:
            self._dep_graph.add_node(kinda_definition)
            self.codeloc_uses.add(kinda_definition)
            self.live_definitions.uses_by_codeloc[self.codeloc].add(kinda_definition)

    def downsize(self):
        self.all_definitions = set()
        self.live_definitions.reset_uses()

    @overload
    def deref(
        self,
        pointer: int | claripy.ast.BV | HeapAddress | SpOffset,
        size: int | DerefSize,
        endness: archinfo.Endness = ...,
    ) -> MemoryLocation | None: ...

    @overload
    def deref(
        self,
        pointer: MultiValues | A | Definition | Iterable[A] | Iterable[Definition[A]],
        size: int | DerefSize,
        endness: archinfo.Endness = ...,
    ) -> set[MemoryLocation]: ...

    def deref(
        self,
        pointer: (
            MultiValues[claripy.ast.BV]
            | Atom
            | Definition[Atom]
            | Iterable[Atom]
            | Iterable[Definition[Atom]]
            | int
            | claripy.ast.BV
            | HeapAddress
            | SpOffset
        ),
        size: int | DerefSize,
        endness: archinfo.Endness = archinfo.Endness.BE,
    ):
        return self.live_definitions.deref(pointer, size, endness)



from __future__ import annotations
from typing import Any, TYPE_CHECKING
from collections import defaultdict
import logging

from angr.ailment import AILBlockRewriter, AILBlockWalker, Const
from angr.ailment.statement import Assignment, SideEffectStatement
from angr.ailment.expression import Atom, Call, VirtualVariable, Convert, BinaryOp, Phi, Extract
from angr.ailment.utils import is_none_or_likeable

from angr.knowledge_plugins.key_definitions import atoms
from angr.code_location import AILCodeLocation

if TYPE_CHECKING:
    from angr.ailment.expression import (
        Expression,
        Load,
        UnaryOp,
        ITE,
        DirtyExpression,
        VEXCCallExpression,
    )
    from angr.ailment.statement import Statement
    from angr.ailment.block import Block


_l = logging.getLogger(__name__)


class ExprNarrowingInfo:
    """
    Stores the analysis result of _narrowing_needed().
    """

    __slots__ = ("narrowable", "phi_vars", "to_size", "use_exprs")

    def __init__(
        self,
        narrowable: bool,
        to_size: int | None = None,
        use_exprs: list[tuple[atoms.VirtualVariable, AILCodeLocation]] | None = None,
        phi_vars: set[VirtualVariable] | None = None,
    ):
        self.narrowable = narrowable
        self.to_size = to_size
        self.use_exprs = use_exprs
        self.phi_vars = phi_vars

    def __repr__(self):
        if self.narrowable:
            return f"<{self.use_exprs} -> {self.to_size} bytes, phi_vars={self.phi_vars}>"
        return "<not narrowable>"


class EffectiveSizeExtractor(AILBlockWalker[None, None, None]):
    """
    Walks a statement or an expression and extracts the effective size (in bits).

    For example, for target expression rax, `(rax & 0xff) + 0x1` means the effective size of rax is 8 bits, from bit
    0 to bit 7. We record this information in expr_to_effective_bits as {rax: (0, 8)}.

    We pay special consideration to expressions that are used as Call arguments, as they may have been converted to a
    smaller size because the Call argument needs that size, but the Call prototype may have been incorrectly inferred.
    """

    def __init__(self, target_expr: Expression, ignore_call_args: bool = True):
        super().__init__()
        self._target_expr = target_expr
        self._ignore_call_args = ignore_call_args
        self.expr_to_effective_bits: dict[Expression, tuple[int, int]] = {}
        self.expr_used_as_call_arg_effective_bits: tuple[int, int] | None = None
        self.expr_used_as_insert_base: bool = False

    def _update_effective_bits(self, expr, lo_bits: int, hi_bits: int):
        existing = self.expr_to_effective_bits.get(expr)
        if existing is None:
            self.expr_to_effective_bits[expr] = lo_bits, hi_bits
        else:
            self.expr_to_effective_bits[expr] = max(existing[0], lo_bits), min(existing[1], hi_bits)

    def _top(self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None):
        pass

    def _stmt_top(self, stmt_idx: int, stmt: Statement, block: Block | None):
        pass

    def _handle_block_end(self, stmt_results, block: Block):
        pass

    def _handle_expr(
        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Any:
        if is_none_or_likeable(expr, self._target_expr):
            # we are done!
            if expr not in self.expr_to_effective_bits:
                self._update_effective_bits(expr, 0, expr.bits)
            return
        super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Insert(self, expr_idx: int, expr, stmt_idx: int, stmt: Statement | None, block: Block | None):
        # self._handle_expr(0, expr.base, stmt_idx, stmt, block)
        if self._target_expr.likes(expr.base):
            self.expr_used_as_insert_base = True
        self._handle_expr(1, expr.offset, stmt_idx, stmt, block)
        self._handle_expr(2, expr.value, stmt_idx, stmt, block)

    def _handle_Extract(self, expr_idx: int, expr, stmt_idx: int, stmt: Statement | None, block: Block | None):
        if isinstance(expr.offset, Const) and isinstance(expr.offset.value, int):
            self._update_effective_bits(expr.base, expr.offset.value, expr.offset.value + expr.bits)
        self._handle_expr(0, expr.base, stmt_idx, stmt, block)
        self._handle_expr(1, expr.offset, stmt_idx, stmt, block)

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement | None, block: Block | None):
        self._handle_expr(0, expr.addr, stmt_idx, stmt, block)

    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement | None, block: Block | None):
        if expr.args is not None:
            for i, arg in enumerate(expr.args):
                handled = False
                if self._ignore_call_args:
                    if (
                        isinstance(arg, Convert)
                        and arg.to_bits < arg.from_bits
                        and is_none_or_likeable(arg.operand, self._target_expr)
                    ):
                        handled = True
                        self.expr_used_as_call_arg_effective_bits = 0, arg.to_bits
                    if (
                        isinstance(arg, Extract)
                        and isinstance(arg.offset, Const)
                        and is_none_or_likeable(arg.base, self._target_expr)
                    ):
                        handled = True
                        self.expr_used_as_call_arg_effective_bits = arg.offset.value, arg.offset.value + arg.bits

                if not handled:
                    self._handle_expr(i, arg, stmt_idx, stmt, block)

    def _handle_SideEffectStatement(self, stmt_idx: int, stmt: SideEffectStatement, block: Block | None):
        if stmt.expr.args is not None:
            for i, arg in enumerate(stmt.expr.args):
                handled = False
                if self._ignore_call_args:
                    if (
                        isinstance(arg, Convert)
                        and arg.to_bits < arg.from_bits
                        and is_none_or_likeable(arg.operand, self._target_expr)
                    ):
                        handled = True
                        self.expr_used_as_call_arg_effective_bits = 0, arg.to_bits
                    if (
                        isinstance(arg, Extract)
                        and isinstance(arg.offset, Const)
                        and is_none_or_likeable(arg.base, self._target_expr)
                    ):
                        handled = True
                        self.expr_used_as_call_arg_effective_bits = arg.offset.value, arg.offset.value + arg.bits

                if not handled:
                    self._handle_expr(i, arg, stmt_idx, stmt, block)

        if stmt.ret_expr is not None:
            self._handle_expr(0, stmt.ret_expr, stmt_idx, stmt, block)

    def _handle_BinaryOp(
        self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        effective_bits = self.expr_to_effective_bits.get(expr)
        if effective_bits is None:
            effective_bits = 0, expr.bits
        if expr.op == "And" and isinstance(expr.operands[1], Const):
            match expr.operands[1].value:
                case 0xFF:
                    lo_bits, hi_bits = 0, 8
                case 0xFFFF:
                    lo_bits, hi_bits = 0, 16
                case 0xFFFF_FFFF:
                    lo_bits, hi_bits = 0, 32
                case 0xFFFF_FFFF_FFFF_FFFF:
                    lo_bits, hi_bits = 0, 64
                case _:
                    lo_bits, hi_bits = effective_bits

            self._update_effective_bits(expr.operands[0], lo_bits, hi_bits)

        elif expr.op in {"Add", "Sub", "Mul", "Mod", "Xor", "Or", "And"}:
            self._update_effective_bits(expr.operands[0], effective_bits[0], effective_bits[1])
            self._update_effective_bits(expr.operands[1], effective_bits[0], effective_bits[1])
        elif expr.op == "Shl":
            self._update_effective_bits(expr.operands[0], effective_bits[0], effective_bits[1])

        self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
        self._handle_expr(1, expr.operands[1], stmt_idx, stmt, block)

    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None):
        if expr.op == "Reference":
            # we really only need 1 byte of the target variable :)
            pass
        else:
            self._update_effective_bits(expr, 0, expr.bits)
        self._handle_expr(0, expr.operand, stmt_idx, stmt, block)

    def _handle_Convert(self, expr_idx: int, expr: Convert, stmt_idx: int, stmt: Statement | None, block: Block | None):
        effective_bits = self.expr_to_effective_bits.get(expr)
        if effective_bits is None or effective_bits[1] > expr.to_bits:
            effective_bits = 0, expr.to_bits
        self._update_effective_bits(expr.operand, effective_bits[0], effective_bits[1])
        self._handle_expr(expr_idx, expr.operand, stmt_idx, stmt, block)

    def _handle_ITE(self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement | None, block: Block | None):
        self._handle_expr(0, expr.cond, stmt_idx, stmt, block)
        self._handle_expr(1, expr.iftrue, stmt_idx, stmt, block)
        self._handle_expr(2, expr.iffalse, stmt_idx, stmt, block)

    def _handle_DirtyExpression(
        self, expr_idx: int, expr: DirtyExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        if expr.operands:
            for i, op in enumerate(expr.operands):
                self._handle_expr(i, op, stmt_idx, stmt, block)

    def _handle_VEXCCallExpression(
        self, expr_idx: int, expr: VEXCCallExpression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        for idx, operand in enumerate(expr.operands):
            self._handle_expr(idx, operand, stmt_idx, stmt, block)


class ExpressionNarrower(AILBlockRewriter):
    """
    Narrows an expression regardless of whether the expression is a definition or a use.
    """

    def __init__(
        self, project, rd, narrowables, addr2blocks: dict[tuple[int, int | None], Block], new_blocks: dict[Block, Block]
    ):
        super().__init__(update_block=False)

        self.project = project
        self._rd = rd
        self._addr2blocks = addr2blocks
        self._new_blocks = new_blocks

        self.new_vvar_sizes: dict[int, int] = {}
        self.replacement_core_vvars: dict[int, list[VirtualVariable]] = defaultdict(list)
        self.narrowed_any = False

        for def_, narrow_info in narrowables:
            self.new_vvar_sizes[def_.atom.varid] = narrow_info.to_size

    def walk(self, block: Block):
        self.narrowed_any = False
        return super().walk(block)

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None) -> Assignment:
        if isinstance(stmt.src, Phi):
            changed = False

            src_and_vvars = []
            for src, vvar in stmt.src.src_and_vvars:
                if vvar is None:
                    src_and_vvars.append((src, None))
                    continue
                if vvar.varid in self.new_vvar_sizes and self.new_vvar_sizes[vvar.varid] != vvar.size:
                    self.narrowed_any = True
                    changed = True
                    new_var = VirtualVariable(
                        vvar.idx,
                        vvar.varid,
                        self.new_vvar_sizes[vvar.varid] * self.project.arch.byte_width,
                        category=vvar.category,
                        oident=vvar.oident,
                        **vvar.tags,
                    )

                    self.replacement_core_vvars[new_var.varid].append(new_var)
                else:
                    new_var = vvar

                src_and_vvars.append((src, new_var))

            new_src = Phi(stmt.src.idx, stmt.src.bits, src_and_vvars, **stmt.src.tags)

        else:
            new_src = self._handle_expr(1, stmt.src, stmt_idx, stmt, block)
            changed = new_src is not stmt.src

        if isinstance(stmt.dst, VirtualVariable) and stmt.dst.varid in self.new_vvar_sizes:
            changed = True
            new_dst = VirtualVariable(
                stmt.dst.idx,
                stmt.dst.varid,
                self.new_vvar_sizes[stmt.dst.varid] * self.project.arch.byte_width,
                category=stmt.dst.category,
                oident=stmt.dst.oident,
                **stmt.dst.tags,
            )

            self.replacement_core_vvars[new_dst.varid].append(new_dst)

            if isinstance(new_src, Phi):
                new_src.bits = self.new_vvar_sizes[stmt.dst.varid] * self.project.arch.byte_width
            else:
                new_src = Convert(
                    None,
                    stmt.src.bits,
                    self.new_vvar_sizes[stmt.dst.varid] * self.project.arch.byte_width,
                    False,
                    new_src,
                    **new_src.tags,
                )
        else:
            new_dst = self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
            changed |= new_dst is not stmt.dst

        if changed:
            self.narrowed_any = True
            assert isinstance(new_dst, Atom)
            return Assignment(stmt.idx, new_dst, new_src, **stmt.tags)

        return stmt

    def _handle_VirtualVariable(
        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        if expr.varid in self.new_vvar_sizes and self.new_vvar_sizes[expr.varid] != expr.size:
            self.narrowed_any = True
            new_expr = VirtualVariable(
                expr.idx,
                expr.varid,
                self.new_vvar_sizes[expr.varid] * self.project.arch.byte_width,
                category=expr.category,
                oident=expr.oident,
                **expr.tags,
            )

            self.replacement_core_vvars[expr.varid].append(new_expr)

            return Convert(
                None,
                new_expr.bits,
                expr.bits,
                False,
                new_expr,
                **new_expr.tags,
            )
        return expr

    def _handle_SideEffectStatement(
        self, stmt_idx: int, stmt: SideEffectStatement, block: Block | None
    ) -> SideEffectStatement:
        new_stmt = super()._handle_SideEffectStatement(stmt_idx, stmt, block)
        assert isinstance(new_stmt, SideEffectStatement)
        changed = new_stmt is not stmt

        if (
            stmt.ret_expr is not None
            and isinstance(stmt.ret_expr, VirtualVariable)
            and stmt.ret_expr.was_reg
            and stmt.ret_expr.varid in self.new_vvar_sizes
            and stmt.ret_expr.size != self.new_vvar_sizes[stmt.ret_expr.varid]
        ):
            changed = True

            # update reg name
            tags = dict(stmt.ret_expr.tags)
            tags["reg_name"] = self.project.arch.translate_register_name(
                stmt.ret_expr.reg_offset, size=self.new_vvar_sizes[stmt.ret_expr.varid]
            )
            new_ret_expr = VirtualVariable(
                stmt.ret_expr.idx,
                stmt.ret_expr.varid,
                self.new_vvar_sizes[stmt.ret_expr.varid] * self.project.arch.byte_width,
                category=stmt.ret_expr.category,
                oident=stmt.ret_expr.oident,
                **tags,
            )
            self.replacement_core_vvars[new_ret_expr.varid].append(new_ret_expr)
            new_stmt.ret_expr = new_ret_expr

        if changed:
            self.narrowed_any = True
            return new_stmt

        return stmt



from __future__ import annotations
from typing import Any, TYPE_CHECKING
import json

from angr.analyses.decompiler.structured_codegen import DummyStructuredCodeGenerator
from angr.analyses.decompiler.structured_codegen.base import CConstantType
from angr.analyses.decompiler.decompilation_cache import DecompilationCache
from angr.knowledge_plugins import StructuredCodeManager
from angr.angrdb.models import DbStructuredCode

if TYPE_CHECKING:
    from angr.analyses.decompiler.structured_codegen.base import IdentType
    from angr.knowledge_base import KnowledgeBase
    from angr.angrdb.models import DbKnowledgeBase


class ConstFormatsSerializer:
    """
    Serialize/deserialize the constant formats dictionary.
    """

    @staticmethod
    def to_json(const_formats: dict[IdentType, dict[str, bool]]) -> dict[int, dict[str, dict[str, bool]]]:
        d = {}
        for key, value in const_formats.items():
            ins_addr, v_type, v = key
            v_type_str = (
                "i" if v_type == CConstantType.INT.value else "f" if v_type == CConstantType.FLOAT.value else "s"
            )  # str
            if ins_addr not in d:
                d[ins_addr] = {}
            d_key = f"{v_type_str}{v}"
            d[ins_addr][d_key] = value
        return d

    @staticmethod
    def from_json(data: dict[int, dict[str, dict[str, str | bool]]]) -> dict[IdentType, dict[str, bool]]:
        r = {}
        for ins_addr, d_ in data.items():
            for d_key, d in d_.items():
                ch = d_key[0]
                if ch == "i":
                    v_type = CConstantType.INT.value
                    value = int(d_key[1:])
                elif ch == "f":
                    v_type = CConstantType.FLOAT.value
                    value = float(d_key[1:])
                else:  # ch == "s"
                    v_type = CConstantType.STRING.value
                    value = d_key[1:]
                key_tpl: IdentType = int(ins_addr), v_type, value
                r[key_tpl] = {}
                for k, v in d.items():
                    r[key_tpl][k] = v is True or (isinstance(v, str) and v.lower() == "true")
        return r


class StructuredCodeManagerSerializer:
    """
    Serialize/deserialize a structured code manager.
    """

    @staticmethod
    def dump(session, db_kb: DbKnowledgeBase, code_manager: StructuredCodeManager):
        """

        :param session:
        :param db_kb:
        :param code_manager:
        :return:
        """

        # remove all existing stored structured code
        session.query(DbStructuredCode).filter_by(kb=db_kb).delete()

        for key, cache in code_manager.cached.items():
            func_addr, flavor = key

            # TODO: Cache types

            expr_comments = None
            if cache.codegen is not None and cache.codegen.expr_comments:
                expr_comments = json.dumps(cache.codegen.expr_comments).encode("utf-8")

            stmt_comments = None
            if cache.codegen is not None and cache.codegen.stmt_comments:
                stmt_comments = json.dumps(cache.codegen.stmt_comments).encode("utf-8")

            const_formats = None
            if cache.codegen is not None and cache.codegen.const_formats:
                const_formats = json.dumps(ConstFormatsSerializer.to_json(cache.codegen.const_formats)).encode("utf-8")

            ite_exprs = None

            db_code = DbStructuredCode(
                kb=db_kb,
                func_addr=func_addr,
                flavor=flavor,
                expr_comments=expr_comments,
                stmt_comments=stmt_comments,
                const_formats=const_formats,
                ite_exprs=ite_exprs,
                errors="\n\n\n".join(cache.errors),
                # configuration=configuration,
            )
            session.add(db_code)

    @staticmethod
    def dict_strkey_to_intkey(d: dict[str, Any]) -> dict[int, Any]:
        new_d = {}

        for key, value in d.items():
            try:
                new_d[int(key)] = value
            except (ValueError, TypeError):
                continue

        return new_d

    @staticmethod
    def load(session, db_kb: DbKnowledgeBase, kb: KnowledgeBase) -> StructuredCodeManager:
        """

        :param session:
        :param db_kb:
        :param kb:
        :return:                        A loaded structured code manager
        """

        manager = StructuredCodeManager(kb)

        db_code_collection = session.query(DbStructuredCode).filter_by(kb=db_kb)

        for db_code in db_code_collection:
            if not db_code.expr_comments:
                expr_comments = None
            else:
                expr_comments = json.loads(db_code.expr_comments.decode("utf-8"))
                expr_comments = StructuredCodeManagerSerializer.dict_strkey_to_intkey(expr_comments)

            if not db_code.stmt_comments:
                stmt_comments = None
            else:
                stmt_comments = json.loads(db_code.stmt_comments.decode("utf-8"))
                stmt_comments = StructuredCodeManagerSerializer.dict_strkey_to_intkey(stmt_comments)

            const_formats = (
                None
                if not db_code.const_formats
                else ConstFormatsSerializer.from_json(json.loads(db_code.const_formats.decode("utf-8")))
            )

            configuration = None
            dummy_codegen = DummyStructuredCodeGenerator(
                db_code.flavor,
                expr_comments=expr_comments,
                stmt_comments=stmt_comments,
                configuration=configuration,
                const_formats=const_formats,
            )
            cache = DecompilationCache(db_code.func_addr)
            cache.codegen = dummy_codegen
            cache.ite_exprs = set()
            cache.errors = db_code.errors.split("\n\n\n")
            manager[(db_code.func_addr, db_code.flavor)] = cache

        return manager



from __future__ import annotations
from collections import defaultdict
from typing import TYPE_CHECKING

from archinfo.arch_arm import is_arm_arch

from angr.knowledge_plugins.plugin import KnowledgeBasePlugin
from .cfg_model import CFGModel

if TYPE_CHECKING:
    from .cfg_model import CFG_ADDR_TYPES


class CFGManager(KnowledgeBasePlugin):
    """
    This is the CFG manager, it manages CFGs
    """

    def __init__(self, kb):
        super().__init__(kb=kb)

        self.cfgs = {}

    def __repr__(self):
        return f"<CFGManager with {len(self.cfgs)} CFGs>"

    def __contains__(self, ident):
        return ident in self.cfgs

    def __getitem__(self, ident) -> CFGModel:
        if ident not in self.cfgs:
            if self._kb is not None and self._kb._project is not None:
                is_arm = is_arm_arch(self._kb._project.arch)
                cache_limit = self._kb._project.get_function_cache_limit()
                edge_cache_limit = self._kb._project.get_cfg_edge_cache_limit()
            else:
                is_arm = False
                cache_limit = None
                edge_cache_limit = None
            self.cfgs[ident] = CFGModel(
                ident,
                cfg_manager=self,
                is_arm=is_arm,
                cache_limit=cache_limit,
                edge_cache_limit=edge_cache_limit,
            )
        return self.cfgs[ident]

    def __setitem__(self, ident, model):
        self.cfgs[ident] = model

    def new_model(self, prefix, addr_type: CFG_ADDR_TYPES = "int"):
        if prefix not in self.cfgs:
            model = self[prefix]
            model.addr_type = addr_type
            return model

        # find a unique ident
        i = 0
        while True:
            ident = f"{prefix}_{i}"
            if ident not in self.cfgs:
                break
            i += 1
        model = self[ident]
        model.addr_type = addr_type
        return model

    def copy(self):
        cm = CFGManager(self._kb)
        cm.cfgs = {x[0]: x[1].copy() for x in self.cfgs.items()}
        return cm

    def get_most_accurate(self) -> CFGModel | None:
        """
        :return: The most accurate CFG present in the CFGManager, or None if it does not hold any.
        """
        less_accurate_to_most_accurate = ["CFGFast", "CFGEmulated"]

        sorted_cfgs_by_prefix = defaultdict(list)
        for key, cfg_model in self.cfgs.items():
            for prefix in less_accurate_to_most_accurate:
                if key.startswith(prefix):
                    the_prefix = prefix
                    break
            else:
                # not found
                continue

            sorted_cfgs_by_prefix[the_prefix].append((key, cfg_model))

        for key in reversed(less_accurate_to_most_accurate):
            if key in sorted_cfgs_by_prefix:
                lst = sorted(sorted_cfgs_by_prefix[key], key=lambda item: item[0])
                return lst[-1][-1]
        return None

    #
    # Pickling
    #

    def __getstate__(self):
        return {
            "_kb": self._kb,
            "cfgs": self.cfgs,
        }

    def __setstate__(self, state):
        self._kb = state["_kb"]
        self.cfgs = state["cfgs"]
        for cfg_model in self.cfgs.values():
            cfg_model._cfg_manager = self


KnowledgeBasePlugin.register_default("cfgs", CFGManager)



# pylint:disable=too-many-boolean-expressions,global-statement,too-many-positional-arguments
from __future__ import annotations
from typing import TYPE_CHECKING
import logging
from enum import Enum

import archinfo
import pyvex


from angr.blade import Blade
from angr.utils.constants import DEFAULT_STATEMENT
from .resolver import IndirectJumpResolver

if TYPE_CHECKING:
    from angr.block import Block


l = logging.getLogger(name=__name__)

PROFILING = False
HITS_CASE_1, HITS_CASE_2, MISSES = 0, 0, 0


def enable_profiling():
    global PROFILING, HITS_CASE_1, HITS_CASE_2, MISSES

    PROFILING = True
    HITS_CASE_1 = 0
    HITS_CASE_2 = 0
    MISSES = 0


def disable_profiling():
    global PROFILING
    PROFILING = False


class Case2Result(Enum):
    """
    Describes the result of resolving case 2 function calls.
    """

    SUCCESS = 0
    FAILURE = 1
    RESUME = 2


class MipsElfFastResolver(IndirectJumpResolver):
    """
    A timeless indirect jump resolver for R9-based indirect function calls in MIPS ELFs.
    """

    def __init__(self, project):
        super().__init__(project, timeless=True)

    def filter(self, cfg, addr, func_addr, block, jumpkind):
        return isinstance(self.project.arch, (archinfo.ArchMIPS32, archinfo.ArchMIPS64))

    def resolve(  # pylint:disable=unused-argument
        self, cfg, addr, func_addr, block, jumpkind, func_graph_complete: bool = True, **kwargs
    ):
        """
        Wrapper for _resolve that slowly increments the max_depth used by Blade for finding sources
        until we can resolve the addr or we reach the default max_depth

        :param cfg: A CFG instance.
        :param int addr: IRSB address.
        :param int func_addr: The function address.
        :param pyvex.IRSB block: The IRSB.
        :param str jumpkind: The jumpkind.
        :return: If it was resolved and targets alongside it
        :rtype: tuple
        """
        global MISSES

        resolved, resolved_targets = self._resolve(cfg, addr, func_addr, block, jumpkind, max_level=2)
        if resolved:
            return resolved, resolved_targets

        if PROFILING:
            MISSES += 1
        return False, []

    def _resolve(self, cfg, addr, func_addr, block, jumpkind, max_level):  # pylint:disable=unused-argument
        """
        Resolves the indirect jump in MIPS ELF binaries where all external function calls are indexed using gp.

        :param cfg: A CFG instance.
        :param int addr: IRSB address.
        :param int func_addr: The function address.
        :param pyvex.IRSB block: The IRSB.
        :param str jumpkind: The jumpkind.
        :param int max_level: maximum level for Blade to resolve when looking for sources
        :return: If it was resolved and targets alongside it
        :rtype: tuple
        """

        global HITS_CASE_1, HITS_CASE_2

        func = cfg.kb.functions.function(addr=func_addr)
        b = Blade(
            cfg.graph,
            addr,
            -1,
            cfg=cfg,
            project=self.project,
            ignore_sp=True,
            ignore_bp=True,
            ignored_regs=("gp",),
            cross_insn_opt=False,
            stop_at_calls=True,
            max_level=max_level,
            include_imarks=False,
        )

        gp_value = func.info.get("gp", None)

        # see if gp is used on this slice at all
        gp_used = self._is_gp_used_on_slice(self.project, b)
        if gp_used and gp_value is None:
            # this might a special case: gp is only used once in this function, and it can be initialized right
            # before its use site.
            # however, it should have been determined in CFGFast
            # cannot determine the value of gp. quit
            l.warning("Failed to determine value of register gp for function %#x.", func.addr)
            return False, []

        # we support two cases:
        # Case 1. t9 is set in the current block, and jalr $t9 at the end of the same block.
        # Case 2. t9 is set in both predecessor blocks, and jalr $t9 at the end of the current block.

        block_addrs = {block_addr for block_addr, _ in b.slice}
        if len(block_addrs) == 2 and addr in block_addrs:
            first_block_addr = next(iter(block_addrs - {addr}))
            r, target = self._resolve_case_2(first_block_addr, block, func_addr, gp_value, cfg)
            if r == Case2Result.SUCCESS:
                if PROFILING:
                    HITS_CASE_2 += 1
                return True, [target]
            if r == Case2Result.FAILURE:
                return False, []
            # otherwise, we need to resume the analysis

        target = self._resolve_case_1(addr, block, func_addr, gp_value, cfg)
        if target is not None:
            if PROFILING:
                HITS_CASE_1 += 1
            return True, [target]

        # no luck
        return False, []

    def _resolve_case_1(self, addr: int, block: pyvex.IRSB, func_addr: int, gp_value: int, cfg) -> int | None:
        # lift the block again with the correct setting

        inital_regs = [(self.project.arch.registers["t9"][0], self.project.arch.registers["t9"][1], func_addr)]
        if gp_value is not None:
            inital_regs.append((self.project.arch.registers["gp"][0], self.project.arch.registers["gp"][1], gp_value))

        first_irsb = self.project.factory.block(
            addr,
            size=block.size,
            collect_data_refs=False,
            const_prop=True,
            cross_insn_opt=False,
            load_from_ro_regions=True,
            initial_regs=inital_regs,
        ).vex_nostmt

        if not isinstance(first_irsb.next, pyvex.IRExpr.RdTmp):
            return None
        target_tmp = first_irsb.next.tmp
        if first_irsb.const_vals is None:
            return None

        # find the value of the next tmp
        for cv in first_irsb.const_vals:
            if cv.tmp == target_tmp:
                target = cv.value
                if self._is_target_valid(cfg, target):
                    return target
                break

        return None

    def _resolve_case_2(
        self, first_block_addr: int, second_block: pyvex.IRSB, func_addr: int, gp_value: int, cfg
    ) -> tuple[Case2Result, int | None]:
        jump_target_reg = self._get_jump_target_reg(second_block)
        if jump_target_reg is None:
            return Case2Result.FAILURE, None
        last_reg_setting_tmp = self._get_last_reg_setting_tmp(second_block, jump_target_reg)
        if last_reg_setting_tmp is not None:
            # the register (t9) is set in this block - we can resolve the jump target using only the current block
            return Case2Result.RESUME, None

        inital_regs = [(self.project.arch.registers["t9"][0], self.project.arch.registers["t9"][1], func_addr)]
        if gp_value is not None:
            inital_regs.append((self.project.arch.registers["gp"][0], self.project.arch.registers["gp"][1], gp_value))

        # lift the first block again with the correct setting
        first_irsb = self.project.factory.block(
            first_block_addr,
            cross_insn_opt=False,
            collect_data_refs=False,
            const_prop=True,
            load_from_ro_regions=True,
            initial_regs=inital_regs,
        ).vex_nostmt

        last_reg_setting_tmp = self._get_last_reg_setting_tmp(first_irsb, jump_target_reg)
        if last_reg_setting_tmp is None:
            return Case2Result.FAILURE, None

        # find the value of the next tmp
        if first_irsb.const_vals is None:
            return Case2Result.FAILURE, None
        for cv in first_irsb.const_vals:
            if cv.tmp == last_reg_setting_tmp:
                target = cv.value
                if self._is_target_valid(cfg, target):
                    return Case2Result.SUCCESS, target
                break

        return Case2Result.FAILURE, None

    @staticmethod
    def _get_jump_target_reg(block: pyvex.IRSB) -> int | None:
        if block.jumpkind != "Ijk_Call":
            return None
        if not isinstance(block.next, pyvex.IRExpr.RdTmp):
            return None
        next_tmp = block.next.tmp

        for stmt in reversed(block.statements):
            if (
                isinstance(stmt, pyvex.IRStmt.Put)
                and isinstance(stmt.data, pyvex.IRExpr.RdTmp)
                and stmt.data.tmp == next_tmp
            ):
                return stmt.offset
            if (
                isinstance(stmt, pyvex.IRStmt.WrTmp)
                and stmt.tmp == next_tmp
                and isinstance(stmt.data, pyvex.IRExpr.Get)
            ):
                return stmt.data.offset

        return None

    @staticmethod
    def _get_last_reg_setting_tmp(block: pyvex.IRSB, target_reg: int) -> int | None:
        for stmt in reversed(block.statements):
            if isinstance(stmt, pyvex.IRStmt.Put) and stmt.offset == target_reg:
                if isinstance(stmt.data, pyvex.IRExpr.RdTmp):
                    return stmt.data.tmp
                return None

        return None

    @staticmethod
    def _is_gp_used_on_slice(project, b: Blade) -> bool:
        gp_offset = project.arch.registers["gp"][0]
        blocks_on_slice: dict[int, Block] = {}
        for block_addr, block_stmt_idx in b.slice.nodes():
            if block_addr not in blocks_on_slice:
                blocks_on_slice[block_addr] = project.factory.block(block_addr, cross_insn_opt=False)
            block = blocks_on_slice[block_addr]
            if block_stmt_idx == DEFAULT_STATEMENT:
                if isinstance(block.vex.next, pyvex.IRExpr.Get) and block.vex.next.offset == gp_offset:
                    gp_used = True
                    break
            else:
                stmt = block.vex.statements[block_stmt_idx]
                if (
                    isinstance(stmt, pyvex.IRStmt.WrTmp)
                    and isinstance(stmt.data, pyvex.IRExpr.Get)
                    and stmt.data.offset == gp_offset
                ):
                    gp_used = True
                    break
        else:
            gp_used = False

        return gp_used



from __future__ import annotations
from collections.abc import Generator
from collections.abc import Iterator


class InsAddrList:
    """
    A memory-efficient replacement for list[int] that stores instruction addresses as a base address plus a bytestring
    of instruction sizes (one byte per instruction).

    Address reconstruction: addr[i] = base_addr + sum(ins_sizes[0:i])

    For compatibility, this class exposes a list-like interface (iteration, indexing,
    length, membership testing, concatenation, etc.).
    """

    __slots__ = ("_base_addr", "_ins_sizes")

    def __init__(self, base_addr: int = 0, ins_sizes: bytes | bytearray | list[int] = b""):
        self._base_addr: int = base_addr
        if isinstance(ins_sizes, (list, bytearray)):
            self._ins_sizes: bytes = bytes(ins_sizes)
        else:
            self._ins_sizes: bytes = ins_sizes

    @classmethod
    def from_addr_list(cls, addrs) -> InsAddrList:
        """Construct an InsAddrList from a list/sequence of absolute instruction addresses."""
        if not addrs:
            return cls()
        base = addrs[0]
        n = len(addrs)
        if n == 1:
            return cls(base, b"\x00")
        sizes = bytearray(n)
        for i in range(n - 1):
            sizes[i] = addrs[i + 1] - addrs[i]
        # Last instruction size is unknown from addresses alone; store 0
        sizes[n - 1] = 0
        return cls(base, bytes(sizes))

    @property
    def base_addr(self) -> int:
        return self._base_addr

    @property
    def ins_sizes(self) -> bytes:
        return self._ins_sizes

    def __len__(self) -> int:
        return len(self._ins_sizes)

    def __bool__(self) -> bool:
        return len(self._ins_sizes) > 0

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return [self[i] for i in range(*idx.indices(len(self._ins_sizes)))]
        if idx < 0:
            idx += len(self._ins_sizes)
        if idx < 0 or idx >= len(self._ins_sizes):
            raise IndexError(idx)
        return self._base_addr + sum(self._ins_sizes[:idx])

    def __iter__(self) -> Generator[int]:
        addr = self._base_addr
        for size in self._ins_sizes:
            yield addr
            addr += size

    def __reversed__(self) -> Iterator[int]:
        return reversed(list(self))

    def __contains__(self, addr) -> bool:
        if not self._ins_sizes:
            return False
        if addr < self._base_addr:
            return False
        cur = self._base_addr
        for size in self._ins_sizes:
            if cur == addr:
                return True
            cur += size
        return False

    def __eq__(self, other):
        if isinstance(other, InsAddrList):
            return self._base_addr == other._base_addr and self._ins_sizes == other._ins_sizes
        if isinstance(other, list):
            if len(other) != len(self._ins_sizes):
                return False
            return all(a == b for a, b in zip(self, other))
        return NotImplemented

    def __add__(self, other) -> list[int]:
        if isinstance(other, (InsAddrList, list)):
            return list(self) + list(other)
        return NotImplemented

    def __radd__(self, other) -> list[int]:
        if isinstance(other, list):
            return other + list(self)
        return NotImplemented

    def __iadd__(self, other) -> list[int]:
        if isinstance(other, (InsAddrList, list)):
            return list(self) + list(other)
        return NotImplemented

    def __repr__(self) -> str:
        return repr(list(self))

    def __hash__(self):
        raise TypeError("unhashable type: 'InsAddrList'")

    def extend(self, other: InsAddrList) -> None:
        addrs = list(self) + list(other)
        self._ins_sizes = bytes([addrs[i + 1] - addrs[i] for i in range(len(addrs) - 1)] + [0])

    def index(self, value, start=0, stop=None):
        if stop is None:
            stop = len(self._ins_sizes)
        addr = self._base_addr + sum(self._ins_sizes[:start])
        for i in range(start, min(stop, len(self._ins_sizes))):
            if addr == value:
                return i
            addr += self._ins_sizes[i]
        raise ValueError(f"{value} is not in InsAddrList")

    def count(self, value) -> int:
        return sum(1 for a in self if a == value)

    def copy(self) -> InsAddrList:
        return InsAddrList(self._base_addr, self._ins_sizes)

    def __reduce__(self):
        # Pickle as (base_addr, list_of_sizes) for readability and compatibility
        return (InsAddrList, (self._base_addr, list(self._ins_sizes)))



from __future__ import annotations
from .base import BaseStructuredCodeGenerator


class DummyStructuredCodeGenerator(BaseStructuredCodeGenerator):
    """
    A dummy structured code generator that only stores user-specified information.
    """

    def __init__(self, flavor: str, expr_comments=None, stmt_comments=None, configuration=None, const_formats=None):
        super().__init__(flavor)
        self.expr_comments = expr_comments
        self.stmt_comments = stmt_comments
        self.configuration = configuration
        self.const_formats = const_formats



# pylint:disable=too-many-boolean-expressions,unused-argument
from __future__ import annotations
import logging
from collections import defaultdict

from archinfo import Endness

from angr.ailment import Block
from angr.ailment.statement import WeakAssignment
from angr.ailment.expression import VirtualVariable, BinaryOp, Const, Load
from angr.sim_type import SimType, SimTypePointer, SimTypeChar

from .optimization_pass import OptimizationPass, OptimizationPassStage

_l = logging.getLogger(name=__name__)


class EagerStdStringConcatenationPass(OptimizationPass):
    """
    Concatenate multiple constant std::string creation calls into one when possible.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.BEFORE_VARIABLE_RECOVERY
    NAME = "Condense multiple constant std::string creation calls into one when possible"
    DESCRIPTION = __doc__.strip()  # type: ignore

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.analyze()

    def _check(self):
        callgraph = self.kb.functions.callgraph
        if self._func.addr not in callgraph:
            # this is a manually created function - always run this optimization pass
            return True, {}
        callees = set(callgraph.successors(self._func.addr))
        string_construction_calls_found = False
        for callee_addr in callees:
            if not self.kb.functions.contains_addr(callee_addr):
                continue
            callee_func = self.kb.functions.get_by_addr(callee_addr)
            demangled_name = callee_func.demangled_name
            if "std::basic_string" in demangled_name and (
                "operator+" in demangled_name or "operator=" in demangled_name
            ):
                string_construction_calls_found = True
                break

        return string_construction_calls_found, {}

    def _analyze(self, cache=None):
        cfg = self.kb.cfgs.get_most_accurate()
        assert cfg is not None

        # iterate over all blocks and look for the following patterns
        # Pattern 1:
        # 00 | LABEL_4025fd:
        # 01 | vvar_21619{s-128|1b} =W Load(addr=0x4511f0<32>, size=222, endness=Iend_BE)
        #
        # 00 | LABEL_40265d:
        # 01 | vvar_21619{s-128|1b} =W (vvar_21619{s-128|1b} + Load(addr=0x451110<32>, size=222, endness=Iend_BE))
        #
        # Pattern 2:
        #
        # 00 | LABEL_40265d:
        # 01 | vvar_21619{s-128|1b} =W (vvar_21619{s-128|1b} + Load(addr=0x451110<32>, size=222, endness=Iend_BE))
        #
        # 00 | LABEL_40268c:
        # 01 | vvar_21619{s-128|1b} =W (vvar_21619{s-128|1b} + Load(addr=0x451030<32>, size=222, endness=Iend_BE))

        entry_block = self._get_block(self.entry_node_addr[0], idx=self.entry_node_addr[1])
        assert entry_block is not None

        traversed = set()
        queue = [entry_block]
        while queue:
            block = queue.pop(0)
            key = block.addr, block.idx
            if key in traversed:
                continue
            traversed.add(key)

            succs = list(self._graph.successors(block))
            for succ in succs:
                succ_key = succ.addr, succ.idx
                if succ_key in traversed:
                    continue
                sub_traversed, new_exits = self._traverse_linear(succ, cfg)
                traversed |= sub_traversed
                queue.extend(new_exits)

    def _traverse_linear(self, start_block: Block, cfg) -> tuple[set[tuple[int, int | None]], list[Block]]:
        traversed = set()
        last_block = start_block

        str_defs: dict[int, bytes] = {}
        succs = []
        stmts_to_remove: dict[int, list[tuple[tuple[int, int | None], int]]] = defaultdict(list)
        while True:
            key = last_block.addr, last_block.idx
            if key in traversed:
                break
            traversed.add(key)

            succs = list(self._graph.successors(last_block))
            if len(succs) > 1 or not succs:
                # this is where the linear traversal ends
                break

            succ = succs[0]

            # find both patterns
            for stmt_idx, stmt in enumerate(succ.statements):
                if isinstance(stmt, WeakAssignment):
                    # pattern 1
                    is_std_string_assign, assigned_str = self._match_std_string_assignment(stmt, cfg)
                    if is_std_string_assign:
                        assert assigned_str is not None
                        # found a std::string assignment
                        str_defs[stmt.dst.varid] = assigned_str
                        stmts_to_remove[stmt.dst.varid].append(((succ.addr, succ.idx), stmt_idx))
                        continue

                    # pattern 2
                    is_std_string_concat, concatenated_str = self._match_std_string_concatenation(stmt, cfg, str_defs)
                    if is_std_string_concat:
                        assert concatenated_str is not None
                        # found a std::string concatenation
                        str_defs[stmt.dst.varid] = concatenated_str
                        stmts_to_remove[stmt.dst.varid].append(((succ.addr, succ.idx), stmt_idx))
                        continue

            last_block = succ

        # process the blocks
        for varid, stmt_locs in stmts_to_remove.items():
            final_str = str_defs[varid]
            str_id = self.kb.custom_strings.allocate(final_str)

            # replace the very last statement with the final string assignment
            last_stmt_block_loc, last_stmt_idx = stmt_locs[-1]
            old_block = self._get_block(last_stmt_block_loc[0], idx=last_stmt_block_loc[1])
            assert old_block is not None
            block = old_block.copy()
            old_stmt = block.statements[last_stmt_idx]
            block.statements[last_stmt_idx] = WeakAssignment(
                old_stmt.idx,
                old_stmt.dst,
                Load(
                    None,
                    Const(None, None, str_id, self.project.arch.bits, custom_string=True),
                    len(final_str),
                    Endness.BE,
                ),
                **old_stmt.tags,
            )
            self._update_block(old_block, block)

            # remove all other statements
            for stmt_block_loc, stmt_idx in reversed(stmt_locs[:-1]):
                old_block = self._get_block(stmt_block_loc[0], idx=stmt_block_loc[1])
                assert old_block is not None
                block = old_block.copy()
                del block.statements[stmt_idx]
                self._update_block(old_block, block)

        return traversed, succs

    def _match_std_string_assignment(self, stmt: WeakAssignment, cfg) -> tuple[bool, bytes | None]:
        if (
            "type" in stmt.tags
            and "dst" in stmt.tags["type"]
            and "src" in stmt.tags["type"]
            and isinstance(stmt.dst, VirtualVariable)
            and isinstance(stmt.src, Load)
            and isinstance(stmt.src.addr, Const)
        ):
            dst_ty, src_ty = stmt.tags["type"]["dst"], stmt.tags["type"]["src"]
            if (
                self._is_std_string_type(dst_ty)
                and self._is_std_string_type_or_charptr(src_ty)
                and isinstance(stmt.src.addr.value, int)
                and stmt.src.addr.value in cfg.memory_data
                and cfg.memory_data[stmt.src.addr.value_int].sort == "string"
            ):
                s = cfg.memory_data[stmt.src.addr.value_int].content
                return True, s
        return False, None

    def _match_std_string_concatenation(
        self, stmt: WeakAssignment, cfg, str_defs: dict[int, bytes]
    ) -> tuple[bool, bytes | None]:
        if (
            "type" in stmt.tags
            and "dst" in stmt.tags["type"]
            and "src" in stmt.tags["type"]
            and isinstance(stmt.dst, VirtualVariable)
            and isinstance(stmt.src, BinaryOp)
            and stmt.src.op == "Add"
        ):
            dst_ty, src_ty = stmt.tags["type"]["dst"], stmt.tags["type"]["src"]
            if self._is_std_string_type(dst_ty) and self._is_std_string_type_or_charptr(src_ty):
                op0, op1 = stmt.src.operands
                if isinstance(op1, VirtualVariable) and isinstance(op0, Load):
                    op0, op1 = op1, op0
                if (
                    isinstance(op0, VirtualVariable)
                    and op0.varid in str_defs
                    and isinstance(op1, Load)
                    and isinstance(op1.addr, Const)
                    and isinstance(op1.addr.value, int)
                    # is op1 a constant string?
                    and op1.addr.value in cfg.memory_data
                    and cfg.memory_data[op1.addr.value].sort == "string"
                ):
                    op1_str = cfg.memory_data[op1.addr.value].content
                    # is op0 also an std::string?
                    op0_str = str_defs[op0.varid]
                    if op0_str is not None and op1_str is not None:
                        final_str = op0_str + op1_str
                        return True, final_str
        return False, None

    @staticmethod
    def _is_std_string_type(t: SimType) -> bool:
        type_str = t.c_repr().removeprefix("const ").removesuffix("&").strip(" ")
        return type_str == "class std::basic_string<char, struct std::char_traits<char>, class std::allocator<char>>"

    @staticmethod
    def _is_std_string_type_or_charptr(t: SimType) -> bool:
        if isinstance(t, SimTypePointer) and isinstance(t.pts_to, SimTypeChar):
            return True
        return EagerStdStringConcatenationPass._is_std_string_type(t)



from __future__ import annotations
from angr.ailment.expression import Const, Load

from angr import SIM_LIBRARIES
from angr.ailment.expression import Call
from angr.calling_conventions import default_cc
from angr.analyses.decompiler.peephole_optimizations.base import PeepholeOptimizationExprBase
from angr.analyses.decompiler.peephole_optimizations import EXPR_OPTS


class APIObfType1PeepholeOptimizer(PeepholeOptimizationExprBase):
    """
    Integrate type-1 deobfuscated API into decompilation output.
    """

    __slots__ = ()

    NAME = "Simplify Type 1 API obfuscation references"
    expr_classes = (Load,)

    def optimize(self, expr: Load, **kwargs):
        if (
            isinstance(expr.addr, Const)
            and (expr.addr.value in self.kb.obfuscations.type1_deobfuscated_apis)
            and expr.bits == self.project.arch.bits
        ):
            # this is actually a function calling a known API
            # replace it with the actual API and the actual arguments
            _, funcname = self.kb.obfuscations.type1_deobfuscated_apis[expr.addr.value]
            if funcname not in self.kb.functions:
                # assign a new function on-demand
                symbol = self.project.loader.extern_object.make_extern(funcname)
                hook_addr = self.project.hook_symbol(
                    symbol.rebased_addr, SIM_LIBRARIES["linux"][0].get_stub(funcname, self.project.arch)
                )
                func = self.kb.functions.function(addr=hook_addr, name=funcname, create=True)
                func.is_simprocedure = True

                default_cc_kwargs = {}
                if self.project.simos is not None:
                    default_cc_kwargs["platform"] = self.project.simos.name
                default_cc_cls = default_cc(self.project.arch.name, **default_cc_kwargs)
                if default_cc_cls is not None:
                    func.calling_convention = default_cc_cls(self.project.arch)
                func.find_declaration(ignore_binary_name=True)
            else:
                func = self.kb.functions[funcname]
            return Const(expr.idx, None, func.addr, self.project.arch.bits, **expr.tags)
        return None


class APIObfType3PeepholeOptimizer(PeepholeOptimizationExprBase):
    """
    Integrate type-3 deobfuscated APIs (calls with const arguments which return APIs) into the decompilation output.
    """

    __slots__ = ()

    NAME = "Simplify Type 3 API obfuscation references"
    expr_classes = (Call,)

    def optimize(self, expr: Call, **kwargs):
        funcbits = self.kb.obfuscations.type3_deobfuscated_apis.get(expr.tags["ins_addr"], None)
        if funcbits is None:
            return None
        dll, api = funcbits
        if not dll:
            # it's likely the main binary
            callees = list(self.project.kb.functions.get_by_name(api, check_previous_names=True))
            if len(callees) == 1:
                return Const(expr.idx, None, callees[0].addr, expr.bits, **(expr.tags | {"always_propagate": True}))
            return None
        if dll not in self.project.loader.shared_objects:
            return None
        sym = self.project.loader.shared_objects[dll].get_symbol(api)
        return Const(expr.idx, None, sym.rebased_addr, expr.bits, **(expr.tags | {"always_propagate": True}))


EXPR_OPTS.append(APIObfType1PeepholeOptimizer)
EXPR_OPTS.append(APIObfType3PeepholeOptimizer)



from __future__ import annotations

import logging

from .plugin import SimStatePlugin

log = logging.getLogger(name=__name__)


class SimStateEdgeHitmap(SimStatePlugin):
    """
    A state plugin that stores AFL-style edge hitmap coverage data.

    This plugin is used by the Icicle engine to track edge coverage during concrete
    execution. It stores the hitmap as raw bytes, which can be efficiently passed
    to and from the Icicle emulator.

    This plugin is NOT registered as a default plugin - it must be explicitly
    added to states that need edge coverage tracking.
    """

    # Standard AFL hitmap size (64KB)
    HITMAP_SIZE = 65536

    def __init__(self, edge_hitmap: bytes | None = None):
        """
        Initialize the edge hitmap plugin.

        :param edge_hitmap: Initial edge hitmap bytes, or None to start with a zeroed hitmap.
        """
        super().__init__()
        # Initialize with zeroed hitmap if not provided
        # Allow None to be set later (e.g., from Icicle emulator)
        self.edge_hitmap: bytes | None = edge_hitmap if edge_hitmap is not None else bytes(self.HITMAP_SIZE)

    def set_state(self, state):
        pass

    @SimStatePlugin.memo
    def copy(self, memo):  # pylint: disable=unused-argument
        """
        Create a copy of the plugin.

        :param memo: Memoization dictionary to avoid duplicate copies.
        :return: A new SimStateEdgeHitmap instance with a copy of the hitmap.
        """
        # bytes are immutable, so we can share the reference
        return SimStateEdgeHitmap(edge_hitmap=self.edge_hitmap)

    def merge(self, others, merge_conditions, common_ancestor=None):  # pylint: disable=unused-argument
        """
        Merge multiple hitmap plugins.

        For coverage tracking, we OR the hitmaps together to preserve all
        observed edges.

        :param others: Other plugin instances to merge with.
        :param merge_conditions: Symbolic conditions for each merge branch.
        :param common_ancestor: Common ancestor plugin instance.
        :return: True if merge was successful.
        """
        for other in others:
            if len(self.edge_hitmap) == len(other.edge_hitmap):
                # OR the hitmaps together to combine coverage
                self.edge_hitmap = bytes(a | b for a, b in zip(self.edge_hitmap, other.edge_hitmap, strict=True))
            else:
                log.warning("Cannot merge edge hitmaps of different sizes")
        return True

    def widen(self, others):  # pylint: disable=unused-argument
        """
        Widening operation for static analysis.

        :param others: Other plugin instances.
        :return: False, widening is not meaningful for hitmaps.
        """
        log.warning("Widening is not implemented for edge hitmap")
        return False



from __future__ import annotations

import logging

import angr
from angr.sim_options import ABSTRACT_MEMORY

l = logging.getLogger(name=__name__)


class mempcpy(angr.SimProcedure):
    # pylint:disable=arguments-differ, missing-class-docstring

    def run(self, dst_addr, src_addr, limit):
        if not self.state.solver.symbolic(limit):
            conditional_size = self.state.solver.eval(limit)
        else:
            max_memcpy_size = self.state.libc.max_memcpy_size  # type: ignore[reportAttributeAccessIssue]
            max_limit = self.state.solver.max_int(limit)
            min_limit = self.state.solver.min_int(limit)
            conditional_size = min(max_memcpy_size, max(min_limit, max_limit))
            if max_limit > max_memcpy_size and conditional_size < max_limit:
                l.warning(
                    "mempcpy upper bound of %#x outside limit, limiting to %#x instead", max_limit, conditional_size
                )

        l.debug("mempcpy running with conditional_size %#x", conditional_size)

        if conditional_size > 0:
            src_mem = self.state.memory.load(src_addr, conditional_size, endness="Iend_BE")
            if ABSTRACT_MEMORY in self.state.options:
                self.state.memory.store(dst_addr, src_mem, size=conditional_size, endness="Iend_BE")
            else:
                self.state.memory.store(dst_addr, src_mem, size=limit, endness="Iend_BE")

        return dst_addr + limit


class __mempcpy_chk(mempcpy):
    # pylint:disable=arguments-differ, missing-class-docstring
    def run(self, dst_addr, src_addr, limit, _destlen):  # type:ignore[reportIncompatibleMethodOverride]
        return super().run(dst_addr, src_addr, limit)



from __future__ import annotations

__all__ = [
    "AILPurityAnalysis",
    "AILPurityDataSource",
    "AILPurityDataUsage",
    "AILPurityResultType",
]

from .analysis import AILPurityAnalysis
from .engine import (
    ResultType as AILPurityResultType,
    DataSource as AILPurityDataSource,
    DataUsage as AILPurityDataUsage,
)



from __future__ import annotations

import logging
import typing
from abc import abstractmethod, ABCMeta
from collections.abc import Iterable
from typing_extensions import override

import claripy
from angr.engines.successors import SimSuccessors, SuccessorsEngine
from angr.sim_state import SimState

log = logging.getLogger(__name__)


HeavyConcreteState = SimState[int, int]


class ConcreteEngine(SuccessorsEngine, metaclass=ABCMeta):
    """
    ConcreteEngine extends SuccessorsEngine for concrete execution.
    """

    @abstractmethod
    def process_concrete(
        self,
        state: HeavyConcreteState,
        num_inst: int | None = None,
        extra_stop_points: set[int] | None = None,
    ) -> HeavyConcreteState:
        """
        Process the concrete state and return a HeavyState object.

        :param state: The concrete state to process.
        :param num_inst: The maximum number of instructions to execute.
        :param extra_stop_points: A set of additional addresses to stop at.
        :return: A HeavyState object representing the processed state.
        """

    @override
    def process_successors(self, successors: SimSuccessors, *, num_inst: int | None = None, **kwargs: typing.Any):
        extra_stop_points_arg = kwargs.pop("extra_stop_points", None)
        extra_stop_points: set[int] | None = None
        if extra_stop_points_arg is not None:
            extra_stop_points = set(typing.cast(Iterable[int], extra_stop_points_arg))

        if len(kwargs) > 0:
            log.warning("ConcreteEngine.process_successors received unknown kwargs: %s", kwargs)

        # TODO: Properly error here when the state is not a HeavyConcreteState
        # Alternatively, we could make SimSuccessors generic over the state type too
        concrete_state = typing.cast(HeavyConcreteState, self.state)

        concrete_successor = self.process_concrete(
            concrete_state, num_inst=num_inst, extra_stop_points=extra_stop_points
        )
        successors.add_successor(
            concrete_successor,
            concrete_successor.ip,
            claripy.true(),
            concrete_successor.history.jumpkind,
            add_guard=False,
        )
        successors.processed = True


__all__ = ["ConcreteEngine", "HeavyConcreteState"]



from __future__ import annotations
import logging

import claripy

import angr
from angr.state_plugins.sim_action_object import SimActionObject

l = logging.getLogger(name=__name__)


class deallocate(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, addr, length):
        if isinstance(addr, SimActionObject):
            addr = addr.ast
        if isinstance(length, SimActionObject):
            length = length.ast

        # return code (see deallocate() docs)
        r = claripy.ite_cases(
            (
                (addr % 0x1000 != 0, self.state.cgc.EINVAL),
                (length == 0, self.state.cgc.EINVAL),
                (self.state.cgc.addr_invalid(addr), self.state.cgc.EINVAL),
                (self.state.cgc.addr_invalid(addr + length), self.state.cgc.EINVAL),
            ),
            claripy.BVV(0, self.state.arch.bits),
        )

        if self.state.solver.symbolic(addr):
            l.warning("Concretizing symbolic address passed to deallocate to max_int")

        addr = self.state.solver.max_int(addr)

        # into a page
        page_size = self.state.memory.page_size
        base_page_num = addr // page_size

        if self.state.solver.symbolic(length):
            l.warning("Concretizing symbolic length passed to deallocate to max_int")

        length = self.state.solver.max_int(length)
        aligned_length = ((length + 0xFFF) // 0x1000) * 0x1000

        # only add sinkholes and unmap on success
        if self.state.solver.max_int(r) == 0:
            # shorten length
            allowed_pages = 0
            while (
                allowed_pages * page_size < aligned_length and base_page_num + allowed_pages in self.state.memory._pages
            ):
                allowed_pages += 1

            if allowed_pages == 0:
                return r

            allowed_length = allowed_pages * page_size
            self.state.cgc.add_sinkhole(addr, allowed_length)

            l.debug("Deallocating [%#x, %#x]", addr, addr + allowed_length - 1)
            self.state.memory.unmap_region(addr, allowed_length)

        return r



from __future__ import annotations

from angr.ailment.block_walker import AILBlockViewer
from angr.ailment.expression import VirtualVariable
from angr.code_location import AILCodeLocation


class FindExtraDefs(AILBlockViewer):
    """
    Simple walker that finds any extra defs
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.found: dict[int, tuple[VirtualVariable, AILCodeLocation]] = {}

    def _handle_UnaryOp(self, expr_idx: int, expr, stmt_idx: int, stmt, block):
        assert block is not None
        assert stmt is not None
        if expr.op == "Reference" and expr.tags.get("extra_def", False):
            assert isinstance(expr.operand, VirtualVariable)
            self.found[expr.operand.varid] = (
                expr.operand,
                AILCodeLocation(block.addr, block.idx, stmt_idx, stmt.tags.get("ins_addr", None)),
            )
        super()._handle_UnaryOp(expr_idx, expr, stmt_idx, stmt, block)



from __future__ import annotations

from .engine import SimEngine
from .failure import SimEngineFailure
from .hook import HooksMixin
from .procedure import ProcedureEngine, ProcedureMixin
from .soot import SootMixin
from .successors import SimSuccessors, SuccessorsEngine
from .syscall import SimEngineSyscall
from .unicorn import SimEngineUnicorn
from .vex import HeavyResilienceMixin, HeavyVEXMixin, SimInspectMixin, SuperFastpathMixin, TrackActionsMixin
from .ail import AILMixin


class UberEngine(
    SimEngineFailure,
    SimEngineSyscall,
    HooksMixin,
    SimEngineUnicorn,
    SuperFastpathMixin,
    TrackActionsMixin,
    SimInspectMixin,
    HeavyResilienceMixin,
    SootMixin,
    AILMixin,
    HeavyVEXMixin,
):
    """
    The default execution engine for angr. This engine includes mixins for most
    common functionality in angr, including VEX IR, unicorn, syscall handling,
    and simprocedure handling.

    For some performance-sensitive applications, you may want to create a custom
    engine with only the necessary mixins.
    """


__all__ = [
    "HeavyResilienceMixin",
    "HeavyVEXMixin",
    "HooksMixin",
    "ProcedureEngine",
    "ProcedureMixin",
    "SimEngine",
    "SimEngineFailure",
    "SimEngineSyscall",
    "SimEngineUnicorn",
    "SimInspectMixin",
    "SimSuccessors",
    "SootMixin",
    "SuccessorsEngine",
    "SuperFastpathMixin",
    "TrackActionsMixin",
    "UberEngine",
]


try:
    from .pcode import HeavyPcodeMixin

    class UberEnginePcode(SimEngineFailure, SimEngineSyscall, HooksMixin, HeavyPcodeMixin):  # pylint:disable=abstract-method
        pass

    __all__.append("UberEnginePcode")

except ImportError:
    pass



# pylint:disable=missing-class-docstring,no-self-use
from __future__ import annotations
from angr.ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class AShlConstSubA(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "(a << N) - a => (a * (2 ** N - 1))"
    expr_classes = (BinaryOp,)

    def optimize(self, expr: BinaryOp, **kwargs):
        if (
            expr.op == "Sub"
            and len(expr.operands) == 2
            and isinstance(expr.operands[0], BinaryOp)
            and expr.operands[0].op == "Shl"
            and isinstance(expr.operands[0].operands[1], Const)
            and expr.signed == expr.operands[0].signed
        ):
            a = expr.operands[1]
            if expr.operands[0].operands[0].likes(a):
                N = expr.operands[0].operands[1].value
                return BinaryOp(
                    expr.idx,
                    "Mul",
                    [a, Const(None, None, 2**N - 1, expr.bits, **expr.operands[0].operands[1].tags)],
                    expr.signed,
                    **expr.tags,
                )

        return None



from __future__ import annotations

from collections.abc import Callable

from cachetools import LRUCache


class SmartLRUCache(LRUCache):
    """
    An LRU cache that supports an eviction callback.
    """

    def __init__(self, maxsize, getsizeof=None, evict: Callable | None = None):
        LRUCache.__init__(self, maxsize, getsizeof=getsizeof)
        self._evict = evict

    def popitem(self):
        key, val = LRUCache.popitem(self)
        if self._evict is not None:
            self._evict(key, val)
        return key, val



from __future__ import annotations
from typing import Generic, TypeVar

JobType = TypeVar("JobType")
JobKey = TypeVar("JobKey")


class JobInfo(Generic[JobType, JobKey]):
    """
    Stores information of each job.
    """

    def __init__(self, key: JobKey, job: JobType):
        self.key = key
        self.jobs = [(job, "")]

        self.narrowing_count = 0  # not used

    def __hash__(self):
        return hash((self.key, *tuple(self.jobs)))

    def __eq__(self, o):
        return type(self) is type(o) and self.key == o.key and self.job == o.job

    def __repr__(self):
        return f"<JobInfo {self.key!s}>"

    @property
    def job(self) -> JobType:
        """
        Get the latest available job.

        :return: The latest available job.
        """

        job, _ = self.jobs[-1]
        return job

    @property
    def merged_jobs(self):
        for job, job_type in self.jobs:
            if job_type == "merged":
                yield job

    @property
    def widened_jobs(self):
        for job, job_type in self.jobs:
            if job_type == "widened":
                yield job

    def add_job(self, job, merged=False, widened=False):
        """
        Appended a new job to this JobInfo node.
        :param job: The new job to append.
        :param bool merged: Whether it is a merged job or not.
        :param bool widened: Whether it is a widened job or not.
        """

        job_type = ""
        if merged:
            job_type = "merged"
        elif widened:
            job_type = "widened"
        self.jobs.append((job, job_type))



from __future__ import annotations
import functools
import os
import sys
import contextlib
from collections import defaultdict
from inspect import Signature
from typing import TYPE_CHECKING, TypeVar, Generic, cast, Any
from collections.abc import Callable
from types import NoneType
from itertools import chain
from traceback import format_exception

import logging
import time
import typing

import psutil

from rich import progress

from angr.misc.plugins import PluginVendor, VendorPreset
from angr.misc import telemetry
from angr.misc.testing import is_testing

if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase
    from angr.project import Project
    from typing_extensions import ParamSpec
    from .identifier import Identifier
    from .callee_cleanup_finder import CalleeCleanupFinder
    from .vsa_ddg import VSA_DDG
    from .cdg import CDG
    from .bindiff import BinDiff
    from .cfg import CFGEmulated
    from .cfg import CFBlanket
    from .cfg import CFG
    from .cfg import CFGFast
    from .static_hooker import StaticHooker
    from .ddg import DDG
    from .congruency_check import CongruencyCheck
    from .reassembler import Reassembler
    from .backward_slice import BackwardSlice
    from .binary_optimizer import BinaryOptimizer
    from .vfg import VFG
    from .loopfinder import LoopFinder
    from .disassembly import Disassembly
    from .veritesting import Veritesting
    from .code_tagging import CodeTagging
    from .boyscout import BoyScout
    from .variable_recovery import VariableRecoveryFast
    from .variable_recovery import VariableRecovery
    from .reaching_definitions import ReachingDefinitionsAnalysis
    from .complete_calling_conventions import CompleteCallingConventionsAnalysis
    from .decompiler.clinic import Clinic
    from .propagator import PropagatorAnalysis
    from .calling_convention import CallingConventionAnalysis
    from .decompiler.decompiler import Decompiler
    from .xrefs import XRefsAnalysis

    AnalysisParams = ParamSpec("AnalysisParams")

l = logging.getLogger(name=__name__)
t = telemetry.get_tracer(name=__name__)


class AnalysisLogEntry:
    def __init__(self, message, exc_info=False):
        if exc_info:
            e_type, value, traceback = sys.exc_info()
            self.exc_type = e_type
            self.exc_value = value
            self.exc_traceback = traceback
        else:
            self.exc_type = None
            self.exc_value = None
            self.exc_traceback = None

        self.message = message

    def format(self) -> str:
        if self.exc_traceback is None:
            return self.message
        return "\n".join((*format_exception(self.exc_type, self.exc_value, self.exc_traceback), "", self.message))

    def __getstate__(self):
        return (
            str(self.__dict__.get("exc_type")),
            str(self.__dict__.get("exc_value")),
            str(self.__dict__.get("exc_traceback")),
            self.message,
        )

    def __setstate__(self, s):
        self.exc_type, self.exc_value, self.exc_traceback, self.message = s

    def __repr__(self):
        if self.exc_type is None:
            msg_str = repr(self.message)
            if len(msg_str) > 70:
                msg_str = msg_str[:66] + "..."
                if msg_str[0] in ('"', "'"):
                    msg_str += msg_str[0]
            return f"<AnalysisLogEntry {msg_str}>"
        msg_str = repr(self.message)
        if len(msg_str) > 40:
            msg_str = msg_str[:36] + "..."
            if msg_str[0] in ('"', "'"):
                msg_str += msg_str[0]
        return f"<AnalysisLogEntry {msg_str} with {self.exc_type.__name__}: {self.exc_value}>"


A = TypeVar("A", bound="Analysis")


class AnalysesHub(PluginVendor[Any]):
    """
    This class contains functions for all the registered and runnable analyses,
    """

    def __init__(self, project):
        super().__init__()
        self.project = project

    def _init_plugin(self, plugin_cls: type[A]) -> AnalysisFactory[A]:
        return AnalysisFactory(self.project, plugin_cls)

    def __getstate__(self):  # type: ignore[reportIncompatibleMethodOverride]
        s = super().__getstate__()
        return (s, self.project)

    def __setstate__(self, sd):
        s, self.project = sd
        super().__setstate__(s)

    def __getitem__(self, plugin_cls: type[A]) -> AnalysisFactory[A]:
        return AnalysisFactory(self.project, plugin_cls)


class KnownAnalysesPlugin(typing.Protocol):
    Identifier: type[Identifier]
    CalleeCleanupFinder: type[CalleeCleanupFinder]
    VSA_DDG: type[VSA_DDG]
    CDG: type[CDG]
    BinDiff: type[BinDiff]
    CFGEmulated: type[CFGEmulated]
    CFB: type[CFBlanket]
    CFBlanket: type[CFBlanket]
    CFG: type[CFG]
    CFGFast: type[CFGFast]
    StaticHooker: type[StaticHooker]
    DDG: type[DDG]
    CongruencyCheck: type[CongruencyCheck]
    Reassembler: type[Reassembler]
    BackwardSlice: type[BackwardSlice]
    BinaryOptimizer: type[BinaryOptimizer]
    VFG: type[VFG]
    LoopFinder: type[LoopFinder]
    Disassembly: type[Disassembly]
    Veritesting: type[Veritesting]
    CodeTagging: type[CodeTagging]
    BoyScout: type[BoyScout]
    VariableRecoveryFast: type[VariableRecoveryFast]
    VariableRecovery: type[VariableRecovery]
    ReachingDefinitions: type[ReachingDefinitionsAnalysis]
    CompleteCallingConventions: type[CompleteCallingConventionsAnalysis]
    Clinic: type[Clinic]
    Propagator: type[PropagatorAnalysis]
    CallingConvention: type[CallingConventionAnalysis]
    Decompiler: type[Decompiler]
    XRefs: type[XRefsAnalysis]


class AnalysesHubWithDefault(AnalysesHub, KnownAnalysesPlugin):
    """
    This class has type-hinting for all built-in analyses plugin
    """


class AnalysisFactory(Generic[A]):
    def __init__(self, project: Project, analysis_cls: type[A]):
        self._project = project
        self._analysis_cls = analysis_cls
        self.__doc__ = ""
        self.__doc__ += analysis_cls.__doc__ or ""
        self.__doc__ += analysis_cls.__init__.__doc__ or ""
        self.__call__.__func__.__signature__ = Signature.from_callable(analysis_cls.__init__)

    def prep(
        self,
        fail_fast: bool | None = None,
        kb: KnowledgeBase | None = None,
        progress_callback: Callable | None = None,
        show_progressbar: bool = False,
    ) -> type[A]:
        if fail_fast is None:
            fail_fast = is_testing

        @functools.wraps(self._analysis_cls.__init__)
        @t.start_as_current_span(self._analysis_cls.__name__)
        def wrapper(*args, **kwargs):
            span = telemetry.get_current_span()
            sig = cast(Signature, self.__call__.__func__.__signature__)
            bound = sig.bind(None, *args, **kwargs)
            for name, val in chain(bound.arguments.items(), bound.arguments.get("kwargs", {}).items()):
                if name in ("kwargs", "self"):
                    continue
                if isinstance(val, (str, bytes, bool, int, float, NoneType)):
                    if val is None:
                        span.set_attribute(f"arg.{name}.is_none", True)
                    else:
                        span.set_attribute(f"arg.{name}", val)
                elif isinstance(val, (list, tuple, set, frozenset)):
                    listval = list(val)
                    if not listval or (
                        isinstance(listval[0], (str, bytes, bool, int, float))
                        and all(type(sval) is type(listval[0]) for sval in listval)
                    ):
                        span.set_attribute(f"arg.{name}", listval)
                elif isinstance(val, dict):
                    listval_keys = list(val)
                    listval_values = list(val.values())
                    if not listval_keys or (
                        isinstance(listval_keys[0], (str, bytes, bool, int, float))
                        and all(type(sval) is type(listval_keys[0]) for sval in listval_keys)
                    ):
                        span.set_attribute(f"arg.{name}.keys", listval_keys)
                    if not listval_values or (
                        isinstance(listval_values[0], (str, bytes, bool, int, float))
                        and all(type(sval) is type(listval_values[0]) for sval in listval_values)
                    ):
                        span.set_attribute(f"arg.{name}.values", listval_values)
                else:
                    span.set_attribute(f"arg.{name}.unrepresentable", True)
            if self._project.filename is not None:
                span.set_attribute("project.binary_name", self._project.filename)
            span.set_attribute("project.arch_name", self._project.arch.name)

            oself = object.__new__(self._analysis_cls)
            oself.named_errors = defaultdict(list)
            oself.errors = []
            oself.log = []

            oself._fail_fast = fail_fast
            oself._name = self._analysis_cls.__name__
            oself.project = self._project
            oself.kb = kb or self._project.kb
            oself._progress_callback = progress_callback

            oself._show_progressbar = show_progressbar
            oself.__init__(*args, **kwargs)
            return oself

        return wrapper  # type: ignore

    def __call__(self, *args, **kwargs) -> A:
        fail_fast = kwargs.pop("fail_fast", is_testing)
        kb = kwargs.pop("kb", self._project.kb)
        progress_callback = kwargs.pop("progress_callback", None)
        show_progressbar = kwargs.pop("show_progressbar", False)

        w = self.prep(
            fail_fast=fail_fast, kb=kb, progress_callback=progress_callback, show_progressbar=show_progressbar
        )

        r = w(*args, **kwargs)
        # clean up so that it's always pickleable
        r._progressbar = None
        return r


class Analysis:
    """
    This class represents an analysis on the program.

    :ivar project:  The project for this analysis.
    :type project:  angr.Project
    :ivar KnowledgeBase kb: The knowledgebase object.
    :ivar _progress_callback: A callback function for receiving the progress of this analysis. It only takes
                                        one argument, which is a float number from 0.0 to 100.0 indicating the current
                                        progress.
    :ivar bool _show_progressbar: If a progressbar should be shown during the analysis. It's independent from
                                    _progress_callback.
    :ivar progress.Progress _progressbar: The progress bar object.
    """

    project: Project
    kb: KnowledgeBase
    _fail_fast: bool
    _name: str
    errors: list[AnalysisLogEntry] = []
    named_errors: defaultdict[str, list[AnalysisLogEntry]] = defaultdict(list)
    _ram_usage: float | None = None
    _last_ramusage_update: float = 0.0
    _progress_callback = None
    _show_progressbar = False
    _progressbar = None
    _task = None

    _PROGRESS_WIDGETS = [
        progress.TaskProgressColumn(),
        progress.BarColumn(),
        progress.TextColumn("Elapsed:"),
        progress.TimeElapsedColumn(),
        progress.TextColumn("Time:"),
        progress.TimeRemainingColumn(),
        progress.TextColumn("{task.description}"),
    ]

    @contextlib.contextmanager
    def _resilience(self, name=None, exception=Exception):
        try:
            yield
        except exception:  # pylint:disable=broad-except
            if self._fail_fast:
                raise
            else:
                error = AnalysisLogEntry("exception occurred", exc_info=True)
                l.error(
                    "Caught and logged %s with resilience: %s",
                    error.exc_type.__name__,
                    error.exc_value,  # type: ignore
                )
                if name is None:
                    self.errors.append(error)
                else:
                    self.named_errors[name].append(error)

    def _initialize_progressbar(self):
        """
        Initialize the progressbar.
        :return: None
        """

        self._progressbar = progress.Progress(*self._PROGRESS_WIDGETS)
        self._task = self._progressbar.add_task(total=100, description="")

        self._progressbar.start()

    def _update_progress(self, percentage, text=None, **kwargs):
        """
        Update the progress with a percentage, including updating the progressbar as well as calling the progress
        callback.

        :param float percentage:    Percentage of the progressbar. from 0.0 to 100.0.
        :param kwargs:              Other parameters that will be passed to the progress_callback handler.
        :return: None
        """

        if self._show_progressbar:
            if self._progressbar is None:
                self._initialize_progressbar()

            assert self._task is not None
            assert self._progressbar is not None
            self._progressbar.update(self._task, completed=percentage)

            if text is not None and self._progressbar:
                self._progressbar.update(self._task, description=text)

        if self._progress_callback is not None:
            self._progress_callback(percentage, text=text, **kwargs)  # pylint:disable=not-callable

    def _finish_progress(self):
        """
        Mark the progressbar as finished.
        :return: None
        """

        if self._show_progressbar:
            if self._progressbar is None:
                self._initialize_progressbar()
            if self._progressbar is not None:
                assert self._task is not None
                self._progressbar.update(self._task, completed=100)
                self._progressbar.stop()
                self._progressbar = None

        if self._progress_callback is not None:
            self._progress_callback(100.0)  # pylint:disable=not-callable

    @staticmethod
    def _release_gil(ctr, freq, sleep_time=0.001):
        """
        Periodically calls time.sleep() and releases the GIL so other threads (like, GUI threads) have a much better
        chance to be scheduled, and other critical components (like the GUI) can be kept responsiveness.

        This is, of course, a hack before we move all computational intensive tasks to pure C++ implementations.

        :param int ctr:     A number provided by the caller.
        :param int freq:    How frequently time.sleep() should be called. time.sleep() is called when ctr % freq == 0.
        :param sleep_time:  Number (or fraction) of seconds to sleep.
        :return:            None
        """

        if ctr != 0 and ctr % freq == 0:
            time.sleep(sleep_time)

    @property
    def ram_usage(self) -> float:
        """
        Return the current RAM usage of the Python process, in bytes. The value is updated at most once per second.
        """

        if time.time() - self._last_ramusage_update > 1:
            self._last_ramusage_update = time.time()
            proc = psutil.Process(os.getpid())
            meminfo = proc.memory_info()
            self._ram_usage = meminfo.rss
        return self._ram_usage if self._ram_usage is not None else -0.1

    def __getstate__(self):
        d = dict(self.__dict__)
        d.pop("_progressbar", None)
        d.pop("_progress_callback", None)
        d.pop("_statusbar", None)
        return d

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __repr__(self):
        return f"<{self._name} Analysis Result at {id(self):#x}>"


default_analyses = VendorPreset()
AnalysesHub.register_preset("default", default_analyses)


def register_analysis(cls, name):
    AnalysesHub.register_default(name, cls)



from __future__ import annotations

# pylint: disable=arguments-differ,arguments-renamed

from typing import Any, TypeAlias
from collections.abc import Callable, MutableMapping
from collections import defaultdict
from dataclasses import dataclass, field
from typing_extensions import Self

import angr
from angr.engines.light import SimEngineLightAIL
from angr import ailment
from angr.knowledge_plugins.functions.function import Function


@dataclass
class DataUsage:
    """
    Facts about a given data source
    """

    ptr_load: bool = False
    ptr_store: bool = False

    def __or__(self, other):
        return DataUsage(
            ptr_load=self.ptr_load | other.ptr_load,
            ptr_store=self.ptr_store | other.ptr_store,
        )

    def __ior__(self, other):
        self.ptr_load |= other.ptr_load
        self.ptr_store |= other.ptr_store
        return self


@dataclass(frozen=True)
class DataSource:
    """
    Descriptor of a given data source: where did this value come from?
    """

    constant_value: int | None = None
    function_arg: int | None = None
    callee_return: Function | str | None = None
    reference_to: int | None = None

    @property
    def callee_return_name(self) -> str | None:
        if self.callee_return is None:
            return None
        if isinstance(self.callee_return, Function):
            return self.callee_return.name
        return self.callee_return


@dataclass
class StateType:
    """
    Internal state for purity analysis
    """

    addr: tuple[int, int | None]
    vars: MutableMapping[int, DataType_co]

    def copy(self):
        return StateType(
            addr=self.addr,
            vars=defaultdict(frozenset, self.vars),
        )


@dataclass
class ResultType:
    """
    Result of purity analysis - how was each data source used and what function calls arguments were passed?
    """

    uses: MutableMapping[DataSource, DataUsage] = field(default_factory=lambda: defaultdict(DataUsage))
    # keyed as: block addr, block idx, stmt idx, call target, call arg idx
    call_args: MutableMapping[tuple[int, int | None, int, Function | str | None, int], DataType_co] = field(
        default_factory=lambda: defaultdict(frozenset)
    )
    ret_vals: MutableMapping[int, DataType_co] = field(default_factory=lambda: defaultdict(frozenset))
    other_storage: MutableMapping[DataSource, DataType_co] = field(default_factory=lambda: defaultdict(frozenset))

    def update(self, other: Self):
        for src, use in other.uses.items():
            self.uses[src] |= use
        for arg, val in other.call_args.items():
            self.call_args[arg] |= val
        for arg, val in other.ret_vals.items():
            self.ret_vals[arg] |= val
        for arg, val in other.other_storage.items():
            self.other_storage[arg] |= val

    def is_pure(
        self,
        pure_functions: set[str] | None = None,
        allow_read_arguments: bool = True,
        allow_write_arguments: bool = False,
        allow_read_globals: bool = True,
        allow_write_globals: bool = False,
    ) -> bool:
        for loc, use in self.uses.items():
            if loc.constant_value is not None and use.ptr_store and not allow_write_globals:
                return False
            if loc.constant_value is not None and use.ptr_load and not allow_read_globals:
                return False
            if loc.function_arg is not None and use.ptr_store and not allow_write_arguments:
                return False
            if loc.function_arg is not None and use.ptr_load and not allow_read_arguments:
                return False
            if loc.callee_return is not None and (
                pure_functions is None or loc.callee_return_name not in pure_functions
            ):
                return False
        for (_, _, _, func, _), _ in self.call_args.items():
            if pure_functions is None:
                return False
            if isinstance(func, Function) and func.name not in pure_functions:
                return False
            if func not in pure_functions:
                return False
        return True


DataType_co: TypeAlias = frozenset[DataSource]
StmtDataType: TypeAlias = None


class PurityEngineAIL(SimEngineLightAIL[StateType, DataType_co, StmtDataType, ResultType]):
    """
    Core of this analysis: THIS data source has THESE uses.

    A use is arithmetic, store, or load.
    """

    def __init__(
        self, project: angr.Project, clinic: angr.analyses.decompiler.Clinic, recurse: Callable[[Function], ResultType]
    ):
        self.clinic = clinic
        self.result = ResultType()
        self.recurse = recurse
        super().__init__(project)

    def initial_state(self, node: ailment.Block):
        assert self.clinic.arg_vvars is not None
        return StateType(
            addr=(node.addr, node.idx),
            vars=defaultdict(
                frozenset,
                (
                    (vvar.varid, frozenset((DataSource(function_arg=idx),)))
                    for idx, (vvar, _) in self.clinic.arg_vvars.items()
                ),
            ),
        )

    def _top(self, bits):
        return frozenset()

    def _is_top(self, expr):
        return expr == frozenset()

    def _expr_single(self, expr: ailment.Expression) -> DataSource:
        result = self._expr(expr)
        if len(result) == 1:
            return next(iter(result))
        return DataSource()

    def _expr_noconst(self, expr: ailment.Expression) -> DataType_co:
        return frozenset(x for x in self._expr(expr) if x.constant_value is None)

    def process(
        self, state: StateType, *, block: ailment.Block | None = None, whitelist: set[int] | None = None, **kwargs
    ) -> ResultType:
        self.tmps = {}
        self.result = ResultType()
        return super().process(state, block=block, whitelist=whitelist, **kwargs)

    def _process_block_end(
        self, block: ailment.Block, stmt_data: list[StmtDataType], whitelist: set[int] | None
    ) -> ResultType:
        return self.result

    def _do_assign(self, dst: ailment.Expression, val: DataType_co):
        match dst:
            case ailment.expression.VirtualVariable():
                self.state.vars[dst.varid] = val
            case ailment.expression.Tmp():
                self.tmps[dst.tmp_idx] = val
            case _:
                raise NotImplementedError

    def _handle_stmt_Assignment(self, stmt: ailment.statement.Assignment) -> StmtDataType:
        if isinstance(stmt.src, ailment.expression.Phi):
            # handled by the analysis layer
            return
        val = self._expr(stmt.src)
        self._do_assign(stmt.dst, val)

    def _handle_stmt_CAS(self, stmt: ailment.statement.CAS) -> StmtDataType:
        raise NotImplementedError

    def _handle_stmt_WeakAssignment(self, stmt: ailment.statement.WeakAssignment) -> StmtDataType:
        raise NotImplementedError

    def _do_store(self, ptr: DataType_co, val: DataType_co):
        for src in ptr:
            if src.reference_to is not None:
                self.state.vars[src.reference_to] = val
            elif self._is_valid_pointer(src):
                self.result.uses[src].ptr_store = True
                if val:
                    self.result.other_storage[src] |= val

    def _handle_stmt_Store(self, stmt: ailment.statement.Store) -> StmtDataType:
        val = self._expr(stmt.data)
        ptr = self._expr(stmt.addr)
        self._do_store(ptr, val)

    def _handle_stmt_Jump(self, stmt: ailment.statement.Jump) -> StmtDataType:
        self._expr_single(stmt.target)

    def _handle_stmt_ConditionalJump(self, stmt: ailment.statement.ConditionalJump) -> StmtDataType:
        self._expr(stmt.condition)
        if stmt.true_target is not None:
            self._expr_single(stmt.true_target)
        if stmt.false_target is not None:
            self._expr_single(stmt.false_target)

    def _handle_stmt_SideEffectStatement(self, stmt: ailment.statement.SideEffectStatement) -> StmtDataType:
        results = self._expr(stmt.expr)
        if stmt.ret_expr is not None:
            assert 0 in results
            self._do_assign(stmt.ret_expr, results[0])
        if stmt.fp_ret_expr is not None:
            assert 0 in results
            self._do_assign(stmt.fp_ret_expr, results[0])

    def _handle_stmt_Return(self, stmt: ailment.statement.Return) -> StmtDataType:
        for i, expr in enumerate(stmt.ret_exprs):
            r = self._expr(expr)
            self.result.ret_vals[i] |= r

    def _handle_stmt_DirtyStatement(self, stmt: ailment.statement.DirtyStatement) -> StmtDataType:
        self._expr(stmt.dirty)

    def _handle_stmt_Label(self, stmt: ailment.statement.Label) -> StmtDataType:
        pass

    def _handle_expr_Extract(self, expr: ailment.expression.Extract) -> DataType_co:
        return frozenset(x for x in self._expr(expr.base) if x.constant_value is not None)

    def _handle_expr_Insert(self, expr: ailment.expression.Insert) -> DataType_co:
        return frozenset(x for x in self._expr(expr.value) if x.constant_value is not None)

    def _handle_expr_Const(self, expr: ailment.expression.Const) -> DataType_co:
        if isinstance(expr.value, int):
            return frozenset((DataSource(constant_value=expr.value),))
        return self._top(expr.bits)

    def _handle_expr_Tmp(self, expr: ailment.expression.Tmp) -> DataType_co:
        return self.tmps[expr.tmp_idx]

    def _handle_expr_VirtualVariable(self, expr: ailment.expression.VirtualVariable) -> DataType_co:
        # allow registers to be uninitialized since callee-save is a thing
        assert (
            self.clinic.function.name == "_security_check_cookie"
            or expr.category == ailment.expression.VirtualVariableCategory.REGISTER
            or expr.varid in self.state.vars
        )
        return self.state.vars[expr.varid]

    def _handle_expr_Phi(self, expr: ailment.expression.Phi) -> DataType_co:
        assert False, "Unreachable"

    def _handle_expr_Convert(self, expr):
        return frozenset(x for x in self._expr(expr.operand) if x.constant_value is not None)

    def _handle_expr_Reinterpret(self, expr: ailment.expression.Reinterpret) -> DataType_co:
        return self._expr(expr.operand)

    def _is_valid_pointer(self, src: DataSource) -> bool:
        return not (
            src.constant_value is not None and self.project.loader.find_object_containing(src.constant_value) is None
        )

    def _do_load(self, ptr: DataType_co) -> DataType_co:
        result: list[DataSource] = []
        for src in ptr:
            if src.reference_to is not None:
                result.extend(self.state.vars[src.reference_to])
            elif self._is_valid_pointer(src):
                self.result.uses[src].ptr_load = True
                if src.constant_value is None:
                    result.append(src)
        return frozenset(result)

    def _handle_expr_Load(self, expr: ailment.expression.Load) -> DataType_co:
        return self._do_load(self._expr(expr.addr))

    def _handle_expr_Register(self, expr: ailment.expression.Register) -> DataType_co:
        return self._top(expr.bits)

    def _handle_expr_ITE(self, expr: ailment.expression.ITE) -> DataType_co:
        self._expr(expr.condition)
        return self._expr(expr.iftrue) | self._expr(expr.iffalse)

    def _do_call(self, expr: ailment.expression.Call) -> MutableMapping[int, DataType_co]:
        args = [self._expr(arg) for arg in expr.args or []]
        seen = None

        if isinstance(expr.target, ailment.Expression):
            target = self._expr_single(expr.target)
            func = None
            if target.constant_value:
                func = self.clinic.project.kb.functions[target.constant_value]
                if not func.is_plt and not func.is_simprocedure:
                    seen = ResultType() if func.name == "_security_check_cookie" else self.recurse(func)
        elif isinstance(expr.target, str):
            # pure functions
            func = expr.target
            seen = None
        else:
            raise TypeError(f"Unexpected call target type {type(expr.target)}")

        if seen is not None:

            def subst(v: DataSource) -> DataType_co:
                if v.function_arg is not None:
                    return args[v.function_arg]
                if v.reference_to is not None:
                    return frozenset()
                return frozenset((v,))

            def substall(v: DataType_co) -> DataType_co:
                result = []
                for vv in v:
                    result.extend(subst(vv))
                return frozenset(result)

            for srcs, val in seen.other_storage.items():
                if not val:
                    continue
                self._do_store(subst(srcs), substall(val))
            for srcs, kind in seen.uses.items():
                for src in subst(srcs):
                    self.result.uses[src] |= kind
            for callsite, vals in seen.call_args.items():
                self.result.call_args[callsite] |= substall(vals)

            return {i: substall(v) for i, v in seen.ret_vals.items()}
        for i, val in enumerate(args):
            self.result.call_args[(self.block.addr, self.block.idx, self.stmt_idx, func, i)] |= val
        # ummmm need to rearrange data model
        return {0: frozenset((DataSource(callee_return=func),))}

    def _handle_expr_Call(self, expr: ailment.expression.Call) -> DataType_co:
        r = self._do_call(expr)
        assert 0 in r
        return r[0]

    def _handle_expr_DirtyExpression(self, expr: ailment.expression.DirtyExpression) -> DataType_co:
        for arg in expr.operands:
            self._expr(arg)
        return self._top(expr.bits)

    def _handle_expr_VEXCCallExpression(self, expr: ailment.expression.VEXCCallExpression) -> DataType_co:
        for arg in expr.operands:
            self._expr(arg)
        return self._top(expr.bits)

    def _handle_expr_MultiStatementExpression(self, expr: ailment.expression.MultiStatementExpression) -> DataType_co:
        for stmt in expr.stmts:
            self._stmt(stmt)
        return self._expr(expr.expr)

    def _handle_expr_BasePointerOffset(self, expr: ailment.expression.BasePointerOffset) -> DataType_co:
        return self._top(expr.bits)

    def _handle_expr_StackBaseOffset(self, expr: ailment.expression.StackBaseOffset) -> DataType_co:
        return self._top(expr.bits)

    @staticmethod
    def __concrete_unop(f: Callable[[Any, int], int]):
        def inner(self, expr: ailment.expression.UnaryOp) -> DataType_co:
            arg = self._expr(expr.operand)
            result = []
            for src in arg:
                if src.constant_value is not None:
                    result.append(DataSource(constant_value=f(self, src.constant_value)))
                # see similar commented code in __concrete_binop
                # else:
                #     result.append(src)
            return frozenset(result)

        return inner

    @__concrete_unop
    def _handle_unop_Not(self, v):
        return int(not v)

    @__concrete_unop
    def _handle_unop_Neg(self, v):
        return -v

    @__concrete_unop
    def _handle_unop_BitwiseNeg(self, v):
        return ~v

    def _handle_unop_Reference(self, expr: ailment.expression.UnaryOp) -> DataType_co:
        assert isinstance(expr.operand, ailment.expression.VirtualVariable)
        return frozenset((DataSource(reference_to=expr.operand.varid),))

    def _handle_unop_Dereference(self, expr: ailment.expression.UnaryOp) -> DataType_co:
        return self._do_load(self._expr(expr.addr))

    def _handle_unop_default(self, expr: ailment.Expression) -> DataType_co:
        return self._expr_noconst(expr.operand)

    _handle_unop_Clz = _handle_unop_default
    _handle_unop_Ctz = _handle_unop_default
    _handle_unop_GetMSBs = _handle_unop_default
    _handle_unop_unpack = _handle_unop_default
    _handle_unop_Sqrt = _handle_unop_default
    _handle_unop_RSqrtEst = _handle_unop_default

    @staticmethod
    def __concrete_binop(f: Callable[[Any, int, int], int | None]):
        def inner(self, expr: ailment.expression.BinaryOp) -> DataType_co:
            arg0 = self._expr(expr.operands[0])
            arg1 = self._expr(expr.operands[1])
            arg0c: list[int] = []
            arg1c: list[int] = []
            result = []
            for arg, argc in ((arg0, arg0c), (arg1, arg1c)):
                for src in arg:
                    if src.constant_value is not None:
                        argc.append(src.constant_value)
                    # this line is weird because it basically means "you can compute however you like with a source
                    # and it will come out with the same taints as before"
                    # preliminary testing indicates this is not desired
                    # else:
                    #     result.append(src)
            if len(arg0c) * len(arg1c) <= 10:  # arbitrary limit
                for c0 in arg0c:
                    for c1 in arg1c:
                        try:
                            m = f(self, c0, c1)
                        except ZeroDivisionError:
                            pass
                        else:
                            if m is None:
                                continue
                            result.append(DataSource(constant_value=m % 2**expr.bits))

            return frozenset(result)

        return inner

    def _handle_binop_Add(self, expr: ailment.expression.BinaryOp) -> DataType_co:
        r = self._handle_binop_Add_basic(expr)  # pylint: disable=no-value-for-parameter
        arg0 = self._expr(expr.operands[0])
        arg1 = self._expr(expr.operands[1])
        return r | arg0 | arg1

    def _handle_binop_Sub(self, expr: ailment.expression.BinaryOp) -> DataType_co:
        r = self._handle_binop_Sub_basic(expr)  # pylint: disable=no-value-for-parameter
        arg0 = self._expr(expr.operands[0])
        return r | arg0

    @__concrete_binop
    def _handle_binop_Add_basic(self, a, b):  # pylint: disable=no-self-use
        return a + b

    @__concrete_binop
    def _handle_binop_Sub_basic(self, a, b):  # pylint: disable=no-self-use
        return a - b

    @__concrete_binop
    def _handle_binop_Mul(self, a, b):
        return a * b

    @__concrete_binop
    def _handle_binop_Div(self, a, b):
        return a // b

    @__concrete_binop
    def _handle_binop_Mod(self, a, b):
        return a % b

    @__concrete_binop
    def _handle_binop_And(self, a, b):
        return a & b

    @__concrete_binop
    def _handle_binop_Or(self, a, b):
        return a | b

    @__concrete_binop
    def _handle_binop_Xor(self, a, b):
        return a ^ b

    @__concrete_binop
    def _handle_binop_Shl(self, a, b):
        return a << b

    @__concrete_binop
    def _handle_binop_Shr(self, a, b):
        return a >> b

    @__concrete_binop
    def _handle_binop_Sar(self, a, b):
        return a >> b

    @__concrete_binop
    def _handle_binop_LogicalAnd(self, a, b):
        return a and b

    @__concrete_binop
    def _handle_binop_LogicalOr(self, a, b):
        return a or b

    @__concrete_binop
    def _handle_binop_CmpEQ(self, a, b):
        return int(a == b)

    @__concrete_binop
    def _handle_binop_CmpNE(self, a, b):
        return int(a != b)

    @__concrete_binop
    def _handle_binop_CmpLT(self, a, b):
        return int(a < b)

    @__concrete_binop
    def _handle_binop_CmpLE(self, a, b):
        return int(a <= b)

    @__concrete_binop
    def _handle_binop_CmpGT(self, a, b):
        return int(a > b)

    @__concrete_binop
    def _handle_binop_CmpGE(self, a, b):
        return int(a >= b)

    @__concrete_binop
    def _handle_binop_Concat(self, a, b):
        if a == 0:
            return b
        return None

    def _handle_binop_default(self, expr: ailment.expression.BinaryOp) -> DataType_co:
        return self._expr_noconst(expr.operands[0]) | self._expr_noconst(expr.operands[1])

    _handle_binop_AddF = _handle_binop_default
    _handle_binop_AddV = _handle_binop_default
    _handle_binop_SubF = _handle_binop_default
    _handle_binop_SubV = _handle_binop_default
    _handle_binop_Mull = _handle_binop_default
    _handle_binop_MulF = _handle_binop_default
    _handle_binop_MulHiV = _handle_binop_default
    _handle_binop_MulV = _handle_binop_default
    _handle_binop_DivF = _handle_binop_default
    _handle_binop_DivV = _handle_binop_default
    _handle_binop_CmpF = _handle_binop_default
    _handle_binop_CmpORD = _handle_binop_default
    _handle_binop_Ror = _handle_binop_default
    _handle_binop_Rol = _handle_binop_default
    _handle_binop_Carry = _handle_binop_default
    _handle_binop_SCarry = _handle_binop_default
    _handle_binop_SBorrow = _handle_binop_default
    _handle_binop_InterleaveLOV = _handle_binop_default
    _handle_binop_InterleaveHIV = _handle_binop_default
    _handle_binop_CasCmpEQ = _handle_binop_default
    _handle_binop_CasCmpNE = _handle_binop_default
    _handle_binop_ExpCmpNE = _handle_binop_default
    _handle_binop_SarNV = _handle_binop_default
    _handle_binop_ShrNV = _handle_binop_default
    _handle_binop_ShlNV = _handle_binop_default
    _handle_binop_CmpEQV = _handle_binop_default
    _handle_binop_CmpNEV = _handle_binop_default
    _handle_binop_CmpGEV = _handle_binop_default
    _handle_binop_CmpGTV = _handle_binop_default
    _handle_binop_CmpLEV = _handle_binop_default
    _handle_binop_CmpLTV = _handle_binop_default
    _handle_binop_MinV = _handle_binop_default
    _handle_binop_MaxV = _handle_binop_default
    _handle_binop_QAddV = _handle_binop_default
    _handle_binop_QNarrowBinV = _handle_binop_default
    _handle_binop_PermV = _handle_binop_default
    _handle_binop_Set = _handle_binop_default



from __future__ import annotations
from angr.procedures.posix.mmap import mmap


class old_mmap(mmap):
    def run(self, ptr):
        addr, length, prot, flags, fd, offset = self.state.mem[ptr].dword.array(6).resolved
        return super().run(addr, length, prot, flags, fd, offset)


class mmap2(mmap):
    def run(self, addr, length, prot, flags, fd, offset):
        if len(offset) == 32:
            offset = offset.zero_extend(32)
            offset *= 0x1000
        return super().run(addr, length, prot, flags, fd, offset)



from __future__ import annotations
import itertools
from angr.ailment.expression import BinaryOp, Const, Convert, Extract, Insert
from .base import PeepholeOptimizationExprBase


class SimplifyBitwiseInserts(PeepholeOptimizationExprBase):
    """
    simplifies _INSERT(A, c, EXTRACT(A, ...) | B) => A | B
    """

    __slots__ = ()

    NAME = "Simplify bitwise inserts"
    expr_classes = (Insert,)

    def optimize(self, expr, **kwargs):
        pb1l = [(0, expr.base)]
        if isinstance(expr.base, BinaryOp) and expr.base.op == "Concat":
            if isinstance(expr.base.operands[0], Const) and expr.base.operands[0].value == 0:
                pb1l.append((expr.base.bits, expr.base.operands[1]))
            else:
                pb1l.append((0, expr.base.operands[0]))
        elif isinstance(expr.base, Convert) and expr.base.operand.bits < expr.base.bits:
            # I don't know if 0 is right here
            pb1l.append((0, expr.base.operand))

        if not isinstance(expr.value, BinaryOp):
            return None
        if expr.value.op not in ("Or", "Xor"):
            return None
        pb2l = []
        for potential_base, other_base in zip(expr.value.operands, reversed(expr.value.operands)):
            pb2l.append((0, potential_base, other_base))
            if isinstance(other_base, Const):
                if isinstance(potential_base, Extract) and isinstance(potential_base.offset, Const):
                    pb2l.append((potential_base.offset, potential_base.base, other_base))
                elif isinstance(potential_base, Convert):
                    # I don't know if 0 is right here
                    pb2l.append((0, potential_base.operand, other_base))

        for (pb1o, pb1), (pb2o, pb2, pb2x) in itertools.product(pb1l, pb2l):  # noqa: B007
            if pb1o == pb2o and pb1.bits == pb2.bits and pb1.likes(pb2):
                break
        else:
            return None

        if pb1o != 0:
            # hello future angr developer! This branch exists because I don't have an example to exercise this branch.
            # for offsets != 0. If you see code enter this branch, please remove this and test for correctness :)
            return None

        if pb2x.bits != pb2.bits:
            # should only be generated by guarded logic above
            assert pb2.bits > pb2x.bits
            assert isinstance(pb2x, Const)
            assert isinstance(pb2x.value, int)
            pb2x = Const(self.manager.next_atom(), None, pb2x.value << pb1o, pb2.bits)
        elif pb1o != 0:
            pb2x = BinaryOp(
                self.manager.next_atom(), "Shl", [pb2x, Const(self.manager.next_atom(), None, pb1o, pb2x.bits)]
            )

        result = BinaryOp(self.manager.next_atom(), expr.value.op, [pb1, pb2x])
        if result.bits != expr.bits:
            result = Convert(self.manager.next_atom(), result.bits, expr.bits, False, result)
        return result



from __future__ import annotations
import logging
import inspect
from typing import Any

import networkx

from angr.ailment import Block
from angr.ailment.statement import ConditionalJump

from .return_duplicator_base import ReturnDuplicatorBase
from .optimization_pass import StructuringOptimizationPass

_l = logging.getLogger(name=__name__)


class ReturnDuplicatorLow(StructuringOptimizationPass, ReturnDuplicatorBase):
    """
    An optimization pass that reverts a subset of Irreducible Statement Condensing (ISC) optimizations, as described
    in the USENIX 2024 paper SAILR. This is the heavy/goto version of the ReturnDuplicator optimization pass.

    Some compilers, including GCC, Clang, and MSVC, apply various optimizations to reduce the number of statements in
    code. These optimizations will take equivalent statements, or a subset of them, and replace them with a single
    copy that is jumped to by gotos -- optimizing for space and sometimes speed.

    This optimization pass will revert those gotos by re-duplicating the condensed blocks. Since Return statements
    are the most common, we use this optimization pass to revert only gotos to return statements. Additionally, we
    perform some additional readability fixups, like not re-duplicating returns to shared components.

    Args:
        func: The function to optimize.
        node_idx_start: The index to start at when creating new nodes. This is used by Clinic to ensure that
            node indices are unique across multiple passes.
        max_opt_iters: The maximum number of optimization iterations to perform.
        max_calls_in_regions: The maximum number of calls that can be in a region. This is used to prevent
            duplicating too much code.
        prevent_new_gotos: If True, this optimization pass will prevent new gotos from being created.
        minimize_copies_for_regions: If True, this optimization pass will minimize the number of copies by doing only
            a single copy for connected in_edges that form a region.
    """

    ARCHES = None
    PLATFORMS = None
    NAME = "Duplicate returns connect with gotos (low)"
    DESCRIPTION = inspect.cleandoc(__doc__[: __doc__.index("Args:")])  # pylint:disable=unsubscriptable-object

    def __init__(
        self,
        *args,
        # settings
        max_opt_iters: int = 4,
        max_calls_in_regions: int = 2,
        prevent_new_gotos: bool = True,
        minimize_copies_for_regions: bool = True,
        region_identifier=None,
        vvar_id_start: int = 0,
        scratch: dict[str, Any] | None = None,
        max_func_blocks: int = 500,
        **kwargs,
    ):
        StructuringOptimizationPass.__init__(
            self,
            *args,
            max_opt_iters=max_opt_iters,
            prevent_new_gotos=prevent_new_gotos,
            require_gotos=True,
            vvar_id_start=vvar_id_start,
            scratch=scratch,
            region_identifier=region_identifier,
            **kwargs,
        )
        ReturnDuplicatorBase.__init__(
            self,
            *args,
            max_calls_in_regions=max_calls_in_regions,
            minimize_copies_for_regions=minimize_copies_for_regions,
            ri=region_identifier,
            vvar_id_start=vvar_id_start,
            scratch=scratch,
            max_func_blocks=max_func_blocks,
        )
        self.analyze()

    def _check(self):
        return ReturnDuplicatorBase._check(self)

    def _should_duplicate_dst(self, src, dst, graph, dst_is_const_ret=False):
        return self._is_goto_edge(src, dst, graph=graph)

    def _is_goto_edge(
        self,
        src: Block,
        dst: Block,
        max_level_check=1,
        *,
        graph: networkx.DiGraph,
    ):
        """
        TODO: Implement a more principled way of checking if an edge is a goto edge with Phoenix's structuring info
        This function only exists because a long-standing bug that sometimes reports the if-stmt addr
        above a goto edge as the goto src.
        """
        # Do a simple and fast check first
        assert self._goto_manager is not None
        is_simple_goto = self._goto_manager.is_goto_edge(src, dst)
        if is_simple_goto:
            return True

        if graph is not None:
            # Special case 1:
            # We need to check for predecessors above the goto and see if they are a goto.
            # This needs to include Jump to deal with loops.
            blocks = [src]
            level_blocks = [src]
            for _ in range(max_level_check):
                new_level_blocks = []
                for lblock in level_blocks:
                    new_level_blocks += list(graph.predecessors(lblock))

                blocks += new_level_blocks
                level_blocks = new_level_blocks

            src_direct_parents = list(graph.predecessors(src))
            for block in blocks:
                if not block or not block.statements:
                    continue

                # special case if-stmts that are next to each other
                if block in src_direct_parents and isinstance(block.statements[-1], ConditionalJump):
                    continue

                if self._goto_manager.is_goto_edge(block, dst):
                    return True

            # Special case 2: A "goto edge" that ReturnDuplicator wants to test might be an edge that Phoenix
            # includes in its loop region (during the cyclic refinement). In fact, Phoenix tends to include as many
            # nodes as possible into the loop region, and generate a goto edge (which ends up in the structured code)
            # from `dst` to the loop successor.
            # an example of this is captured by the test case `TestDecompiler.test_stty_recover_mode_ret_dup_region`.
            # until someone (ideally @mahaloz) implements a more principled way of translating "goto statements" that
            # Phoenix generates and "goto edges" that ReturnDuplicator tests, we rely on the following stopgap to
            # handle this case.
            node = dst
            while True:
                succs = list(graph.successors(node))
                if len(succs) != 1:
                    break
                succ = succs[0]
                if succ is node:
                    # loop!
                    break
                succ_preds = list(graph.predecessors(succ))
                if len(succ_preds) != 1:
                    break
                if self._goto_manager.is_goto_edge(node, succ):
                    return True
                # keep testing the next edge
                node = succ

        return False

    def _analyze(self, cache=None):
        """
        This analysis is run in a loop in analyze() for a maximum of max_opt_iters times.
        """
        return self._analyze_core(self.out_graph)



from __future__ import annotations
import claripy
import logging

from angr.procedures.java import JavaSimProcedure

log = logging.getLogger(name=__name__)


class CharacterIsDigit(JavaSimProcedure):
    __provides__ = (("java.lang.Character", "isDigit(char)"),)

    def run(self, char_ref):
        log.debug(f"Called SimProcedure java.lang.Character.isDigit with args: {char_ref}")
        char_str = self.state.memory.load(char_ref)

        constraint = claripy.StrIsDigit(char_str)

        return claripy.If(constraint, claripy.BVV(1, 32), claripy.BVV(0, 32))


class CharacterIsSpaceChar(JavaSimProcedure):
    __provides__ = (("java.lang.Character", "isSpaceChar(char)"),)

    def run(self, char_ref):
        log.debug(f"Called SimProcedure java.lang.Character.isSpaceChar with args: {char_ref}")
        char_str = self.state.memory.load(char_ref)

        # Should we add other unicode SPACE_SEPARATOR?
        return claripy.If(char_str == " ", claripy.BVV(1, 32), claripy.BVV(0, 32))
